
/**********************************************************************
 * $Id: groupCom.c,v 1.9 92/04/20 16:06:30 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 **********************************************************************/

#include <xerion/commands.h>

/***********************************************************************
 *	Name:		command_addGroup
 *	Description:	adds a group to the current network, automatically
 *			connects bias unit to all units
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_createGroup - 0 on failure, 1 on success
 ***********************************************************************/
int	command_addGroup (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  char	*name ;
  int	numUnits, idx ;
  int	newMask, groupMask ;
  Group	group, bias ;

  IUsage("[ -type TYPE [ -type TYPE ...] ] <name> n") ;
  if (GiveHelp(tokc)) {
    ISynopsis("create a group in the current network") ;
    IHelp
      (IHelpArgs,
       "Creates a group with the given <name>,  and  with the specified type",
       "masks  set.  TYPE may be  INPUT, BIAS,  OUTPUT, HIDDEN, or  any user",
       "created mask.   Multiple types may be specified,  each preceded with",
       "the keyword \"-type\".",
       "",
       "The group is  created in the  current network and contains n  units.",
       "The   name   of the group  must  be   unique. The    bias  unit will",
       "automatically be connected to all units in the group.",
       "EXAMPLE",
       "To add an input group with name \"InputGroup\" and 24 units,",
       "",
       "\txerion-> addGroup -type INPUT InputGroup 24",
       "SEE ALSO",
       "deleteGroups, connectGroups, disconnectGroups",
       NULL);
    return 0;
  }

  if (tokc < 3 || tokc % 2 != 1)
    IErrorAbort(IPrintUsage(tokv[0], usage)) ;

  /* Check for all the possible errors */
  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  groupMask = 0 ;
  for (idx = 1 ; 
       strncmp(tokv[idx], "-type", strlen(tokv[idx])) == 0 ; idx += 2) {
    newMask = findClassMask(GROUP_CLASS, tokv[idx+1]) ;
    if (newMask == 0) {
      IErrorAbort("Unknown group type: \"%s\".", tokv[idx+1]) ;
      return 0 ;
    }
    groupMask |= newMask ;
  }
  name = tokv[idx] ;
  numUnits = atoi(tokv[idx+1]) ;

  if (numUnits < 0)
    IErrorAbort("The group cannot have a negative number of units.") ;

  group = groupFromName(currentNet, name) ;
  if (group != NULL)
    IErrorAbort("Group \"%s\" already exists.", group->name) ;
  
  /* Now create the group and add the units */
  group = createGroup(name, groupMask, currentNet) ;

  if (group == NULL)
    IErrorAbort("Cannot create group \"%s\"", name) ;

  for (idx = 0 ; idx < numUnits ; ++idx) {
    char	unitName[BUFSIZ] ;
    sprintf(unitName, "%s.%d", name, idx) ;
    if (0 == createUnit(unitName, group))
      IErrorAbort("Cannot create unit \"%s\"", unitName) ;
  }    

  if (!(group->type & BIAS)) {
    bias = groupFromName(currentNet, "Bias") ;
    if (bias != NULL)
      connectGroups(bias, group, 0) ;
  }
  markToRebuildDisplay(ACTIVITY_DISPLAY | CONNECTION_DISPLAY) ;

  return 1 ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		command_deleteGroups
 *	Description:	deletes groups from the current network
 *			uses regex to match names
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_deleteGroup - 0 on failure, 1 on success
 ***********************************************************************/
int	command_deleteGroups (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  int	idx ;
  Group	*groupArray, *group ;

  IUsage("<group1> [ <group2> ... ]") ;
  if (GiveHelp(tokc)) {
    ISynopsis("deletes groups from the current network") ;
    IHelp
      (IHelpArgs,
       "Deletes  the groups  with names   <group1>   etc. from the   current",
       "network. Names may contain pattern matching  expressions of the form",
       "used by grep.",
       "EXAMPLES",
       "To delete the group with name \"InputGroup\",",
       "",
       "\txerion-> deleteGroups InputGroup",
       "",
       "To delete all groups whose names begin with \"Hidden\",",
       "",
       "\txerion-> deleteGroups Hidden.*",
       "SEE ALSO",
       "addGroup, connectGroups, disconnectGroups",
       NULL);
    return 0;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  for (idx = 1 ; idx < tokc ; ++idx) {
    groupArray = groupsFromRegex(currentNet, tokv[idx]) ;
    if (groupArray == NULL)
      IErrorAbort("Bad string: \"%s\"", tokv[idx]) ;
    if (groupArray[0] == NULL)
      IError("No match: \"%s\"", tokv[idx]) ;

    for (group = groupArray ; *group != NULL ; ++group) {
      destroyGroup (*group) ;
    }

    free(groupArray) ;
  }

  markToRebuildDisplay(ACTIVITY_DISPLAY | CONNECTION_DISPLAY) ;

  return 1 ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		command_connectGroups
 *	Description:	connects two groups togther completely
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_connectGroup - 0 on failure, 1 on success
 ***********************************************************************/
int	command_connectGroups (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  int	linkMask, newMask, idx ;
  Group	*preGroupArray,  *preGroup ;
  Group	*postGroupArray, *postGroup ;

  IUsage("[ -type TYPE [ -type TYPE ...] ] <group1> <group2>") ;
  if (GiveHelp(tokc)) {
    ISynopsis("connects two sets of groups to each other") ;
    IHelp
      (IHelpArgs,
       "Connects each unit  in <group1>  to all of   the units in  <group2>,",
       "where <group1> and  <group2>  are the names  of two  groups  in  the",
       "current network.  Order   is important.  <group1> is the   preceding",
       "group to <group2>.  Group names may  use regular expressions similar",
       "to those used in grep to connect more  than two groups.  Optionally,",
       "a type mask  may be  specified for the  type of link  to create.  At",
       "present there are no standard link types.",
       "EXAMPLE",
       "\txerion-> connectGroups \"bias\" \"hidden.*\"",
       "",
       "will connect  the group with  name  \"bias\" to all  groups whose name",
       "begins with the string \"hidden\".",
       "SEE ALSO",
       "connectUnits, disconnectGroups, disconnectUnits",
       NULL);
    return 0;
  }

  if (tokc < 3 || tokc % 2 != 1) {
    IErrorAbort(IPrintUsage(tokv[0], usage)) ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  linkMask = 0 ;
  for (idx = 1 ; 
       strncmp(tokv[idx], "-type", strlen(tokv[idx])) == 0 ; idx += 2) {
    newMask = findClassMask(LINK_CLASS, tokv[idx+1]) ;
    if (newMask == 0) {
      IErrorAbort("Unknown type: \"%s\".", tokv[idx+1]) ;
      return 0 ;
    }
    linkMask |= newMask ;
  }

  /* get the list of pre groups assuming tokv[idx] is a regular expression */
  preGroupArray = groupsFromRegex(currentNet, tokv[idx]) ;
  if (preGroupArray == NULL)
    IErrorAbort("Bad string: \"%s\"", tokv[idx]) ;
  if (preGroupArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", tokv[idx]) ;
  ++idx ;

  /* get the list of post groups assuming tokv[idx] is a regular expression */
  postGroupArray = groupsFromRegex(currentNet, tokv[idx]) ;
  if (postGroupArray == NULL)
    IErrorAbort("Bad string: \"%s\"", tokv[idx]) ;
  if (postGroupArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", tokv[idx]) ;

  for (preGroup = preGroupArray ; *preGroup != NULL ; ++preGroup) {
    for (postGroup = postGroupArray ; *postGroup != NULL ; ++postGroup) {
      if (0 == connectGroups(*preGroup, *postGroup, linkMask))
	IErrorAbort("Cannot connect groups \"%s\" and \"%s\"",
		    (*preGroup)->name, (*postGroup)->name) ;
    }
  }

  free(preGroupArray) ;
  free(postGroupArray) ;

  markToRebuildDisplay(CONNECTION_DISPLAY) ;

  return 1 ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		command_disconnectGroups
 *	Description:	disconnects two groups
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_disconnectGroup - 0 on failure, 1 on success
 ***********************************************************************/
int	command_disconnectGroups (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  Group	*preGroupArray,  *preGroup ;
  Group	*postGroupArray, *postGroup ;

  IUsage("<group1> <group2>") ;
  if (GiveHelp(tokc)) {
    ISynopsis("disconnects two sets of groups from each other") ;
    IHelp
      (IHelpArgs,
       "Disconnects each unit in <group1> from all of the units in <group2>,",
       "where  <group1> and <group2>  are  the names  of  two  groups in the",
       "current network.    Order is  important.  <group1>  is the preceding",
       "group to <group2>.  Group  names may use regular expressions similar",
       "to those used in grep to disconnect more than two groups.",
       "EXAMPLE",
       "\txerion-> disconnectGroups \"bias\" \"hidden.*\"",
       "",
       "will  disconnect the group with  name \"bias\" from all   groups whose",
       "name begins with the string \"hidden\".",
       "SEE ALSO",
       "connectGroups, connectUnits, disconnectUnits",
       NULL);
    return 0;
  }

  if (tokc != 3) {
    IErrorAbort(IPrintUsage(tokv[0], usage)) ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  /* get the list of pre groups assuming tokv[1] is a regular expression */
  preGroupArray = groupsFromRegex(currentNet, tokv[1]) ;
  if (preGroupArray == NULL)
    IErrorAbort("Bad string: \"%s\"", tokv[1]) ;
  if (preGroupArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", tokv[1]) ;

  /* get the list of post groups assuming tokv[2] is a regular expression */
  postGroupArray = groupsFromRegex(currentNet, tokv[2]) ;
  if (postGroupArray == NULL)
    IErrorAbort("Bad string: \"%s\"", tokv[2]) ;
  if (postGroupArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", tokv[2]) ;

  for (preGroup = preGroupArray ; *preGroup != NULL ; ++preGroup) {
    for (postGroup = postGroupArray ; *postGroup != NULL ; ++postGroup) {
      disconnectGroups(*preGroup, *postGroup) ;
    }
  }

  free(preGroupArray) ;
  free(postGroupArray) ;

  markToRebuildDisplay(CONNECTION_DISPLAY) ;

  return 1 ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		command_orderGroups
 *	Description:	orders the array of groups in the current net.
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_orderGroups - 0 on failure, 1 on success
 ***********************************************************************/
static int	compare ARGS((const void	*p1, const void	*p2)) ;
static Group	*groupArray ;
static int	numGroups ;
int	command_orderGroups (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  int	idx ;

  IUsage("<group1> [ <group2> ... ]") ;
  if (GiveHelp(tokc)) {
    ISynopsis("orders the groups in the current net") ;
    IHelp
      (IHelpArgs,
       "Sets the  order for the forward  pass in the  current  network.  The",
       "backward  pass uses the  reverse order to the forward  pass.  If any",
       "groups  are not specified, their  position  in the  ordering is  not",
       "certain.",
       "SEE ALSO",
       "connectGroups",
       NULL);
    return 0;
  }

  /* Check for all the possible errors */
  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;
  if (currentNet->numGroups == 0)
    IErrorAbort("There are no groups in the current net.") ;


  ++tokv ;
  --tokc ;
  groupArray = (Group *) callocOrAbort(currentNet->numGroups, sizeof(Group)) ;
  numGroups  = currentNet->numGroups ;
  for (idx = 0 ; idx < tokc ; ++idx) {
    if ((groupArray[idx] = groupFromName(currentNet, tokv[idx])) == NULL)
      IErrorAbort("Unknown group: \"%s\"", tokv[idx]) ;
  }

  qsort(currentNet->group, currentNet->numGroups, sizeof(Group), compare) ;
  free(groupArray) ;
  return 1 ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		groupTypes
 *	Description:	prints the types of the groups in the net
 *	Parameters:
 *	int tokc - 
 *	char *tokv[] -
 *	Return Value:
 *	  int command_groupTypes - 
 *********************************************************************/
static void	printGroupTypes(group, data)
  Group		group ;
  void		*data ;
{
  FILE		*fp = (FILE *)data ;
  Mask		mask;
  int		j;

  fprintf(fp, "Group %s:\t", group->name);
  mask = 1;
  for (j=0; j<8*sizeof(Mask); j++) {
    if (group->type & mask)
      fprintf(fp, " %s", classMaskName(GROUP_CLASS, mask));
    mask <<= 1;
  }
  fprintf(fp, "\n");
}
/**********************************************************************/
int command_groupTypes(tokc, tokv)
int tokc;
char *tokv[];
{
  struct ARG_DESC *argd;
  argd = StartArgs(tokv[0]);
  /* Args(argd, "<name>%s", &name); */ /* for example */
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("prints out the types for all groups");
    IHelp(tokc, tokv[0], NULL, synopsis,
	  "\"groupTypes\" prints  out  the types of  the groups   in the  current",
	  "network. It gets these types from the type mask field of the groups.",
	  "",
	  NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);

  if (currentNet==NULL)
    IErrorAbort("current net is NULL");

  netForAllGroups(currentNet, ALL, printGroupTypes, dout) ;

  return 0;
}
/********************************************************************/



/***********************************************************************
 *	Name:		compare
 *	Description:	compares to pointers to Groups and figures out
 *			which one comes first in groupArray
 *	Parameters:	the pointers to the group
 *	Return Value:	-1, 0, 1 depending on which comes first.
 ***********************************************************************/
static int	compare(p1, p2)
  const void	*p1 ;
  const void	*p2 ;
{
  int	i1, i2 ;
  
  for (i1 = 0 ; i1 < numGroups && *(Group *)p1 != groupArray[i1] ; ++i1) 
    ;

  for (i2 = 0 ; i2 < numGroups && *(Group *)p2 != groupArray[i2] ; ++i2) 
    ;

  return i1 - i2 ;
}
/**********************************************************************/
