
/**********************************************************************
 * $Id: unitCom.c,v 1.16 92/04/20 16:07:03 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 **********************************************************************/

#include <xerion/commands.h>

/***********************************************************************
 *	Name:		command_addUnit
 *	Description:	adds a unit to a group in the current network
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_addUnit - 0 on failure, 1 on success
 ***********************************************************************/
int	command_addUnit (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  char	*name, *groupName ;
  Group	group ;
  Unit	unit, bias ;

  IUsage("<name> <group>") ;
  if (GiveHelp(tokc)) {
    ISynopsis("add a unit to an existing group in the current network") ;
    IHelp
      (IHelpArgs,
       "Creates a unit with name  <name> and adds  it to the group with name",
       "<group> in the current network.  The name must  be unique.  The bias",
       "unit will be automatically connected to the unit.",
       "EXAMPLES",
       "\txerion-> addUnit BiasUnit BiasGroup",
       "SEE ALSO",
       "deleteUnits, connectUnits, disconnectUnits",
       NULL);
    return 0;
  }

  if (tokc != 3)
    IErrorAbort(IPrintUsage(tokv[0], usage)) ;

  name      = tokv[1] ;
  groupName = tokv[2] ;

  /* Check for all the possible errors */
  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  unit = unitFromName(currentNet, name) ;
  if (unit != NULL)
    IErrorAbort("Unit \"%s\" already exists.", unit->name) ;
  
  group = groupFromName(currentNet, groupName) ;
  if (group == NULL)
    IErrorAbort("Group \"%s\" is not in the current net.", groupName) ;

  /* Now create the unit int the requested group */
  unit = createUnit(name, group) ;

  if (unit == NULL)
    IErrorAbort("Cannot create unit \"%s\"", name) ;

  if (!(group->type & BIAS)) {
    bias = unitFromName(currentNet, "Bias") ;
    if (bias != NULL)
      if (connectUnits(bias, unit, 0) == NULL)
	IErrorAbort("Cannot connect \"%s\" to bias", unit->name) ;
  }

  markToRebuildDisplay(ACTIVITY_DISPLAY | CONNECTION_DISPLAY) ;

  return 1 ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		command_deleteUnits
 *	Description:	deletes units from the current network
 *			uses regex to match namess
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_deleteUnit - 0 on failure, 1 on success
 ***********************************************************************/
int	command_deleteUnits (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  int	idx ;
  Unit	*unitArray, *unit ;

  IUsage("<unit1> [ <unit2> ... ]") ;
  if (GiveHelp(tokc)) {
    ISynopsis("deletes units from the current network") ;
    IHelp
      (IHelpArgs,
       "Deletes the units with names <unit1>  etc. from the current network.",
       "Names may contain pattern  matching expressions of  the form used by",
       "grep.",
       "EXAMPLES",
       "To delete the unit with name \"InputUnit\",",
       "",
       "\txerion-> deleteUnits InputUnit",
       "",
       "To delete all units whose names begin with \"Hidden\",",
       "",
       "\txerion-> deleteUnits Hidden.*",
       "SEE ALSO",
       "addUnit, connectUnits, disconnectUnits",
       NULL);
    return 0;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  for (idx = 1 ; idx < tokc ; ++idx) {
    unitArray = unitsFromRegex(currentNet, tokv[idx]) ;
    if (unitArray == NULL)
      IErrorAbort("Bad string: \"%s\"", tokv[idx]) ;
    if (unitArray[0] == NULL)
      IError("No match: \"%s\"", tokv[idx]) ;

    for (unit = unitArray ; *unit != NULL ; ++unit) {
      destroyUnit (*unit) ;
    }

    free(unitArray) ;
  }
  markToRebuildDisplay(ACTIVITY_DISPLAY | CONNECTION_DISPLAY) ;

  return 1 ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		command_connectUnits
 *	Description:	connects two units togther
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_connectUnit - 0 on failure, 1 on success
 ***********************************************************************/
int	command_connectUnits (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  int	linkMask, newMask, idx ;
  Unit	*preUnitArray,  *preUnit ;
  Unit	*postUnitArray, *postUnit ;

  IUsage("[ -type TYPE [ -type TYPE ...] ] <unit1> <unit2>") ;
  if (GiveHelp(tokc)) {
    ISynopsis("connects two sets of units to each other") ;
    IHelp
      (IHelpArgs,
       "Connects <unit1> to <unit2>, where <unit1> and <unit2> are the names",
       "of two units in  the current network.   Order, is important. <unit1>",
       "is  the preceding  unit  to <unit2>.   Unit  names may   use regular",
       "expressions similar  to   grep  to  connect more    than two  units.",
       "Optionally, a  type mask may  be specified for  the type  of link to",
       "create. At present there are no standard link types.",
       "EXAMPLES",
       "\tconnectUnits \"bias\" \"hidden.*\"",
       "",
       "will   connect the unit with  name  \"bias\" to all  units  whose name",
       "begins with the string \"hidden\".",
       "SEE ALSO",
       "connectGroups, disconnectGroups, disconnectUnits",
       NULL);
    return 0;
  }

  if (tokc < 3 || tokc % 2 != 1) {
    IErrorAbort(IPrintUsage(tokv[0], usage)) ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  linkMask = 0 ;
  for (idx = 1 ; 
       strncmp(tokv[idx], "-type", strlen(tokv[idx])) == 0 ; idx += 2) {
    newMask = findClassMask(LINK_CLASS, tokv[idx+1]) ;
    if (newMask == 0) {
      IErrorAbort("Unknown type: \"%s\".", tokv[idx+1]) ;
      return 0 ;
    }
    linkMask |= newMask ;
  }

  /* get the list of pre groups assuming tokv[idx] is a regular expression */
  preUnitArray = unitsFromRegex(currentNet, tokv[idx]) ;

  if (preUnitArray == NULL)
    IErrorAbort("Bad string: \"%s\"", tokv[idx]) ;
  if (preUnitArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", tokv[idx]) ;

  ++idx ;

  /* get the list of post units assuming tokv[idx] is a regular expression */
  postUnitArray = unitsFromRegex(currentNet, tokv[idx]) ;

  if (postUnitArray == NULL)
    IErrorAbort("Bad string: \"%s\"", tokv[idx]) ;
  if (postUnitArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", tokv[idx]) ;

  for (preUnit = preUnitArray ; *preUnit != NULL ; ++preUnit) {
    for (postUnit = postUnitArray ; *postUnit != NULL ; ++postUnit) {
      if (NULL == connectUnits(*preUnit, *postUnit, linkMask))
	IErrorAbort("Cannot connect units \"%s\" and \"%s\"",
		    (*preUnit)->name, (*postUnit)->name) ;
    }
  }

  free(preUnitArray) ;
  free(postUnitArray) ;

  markToRebuildDisplay(CONNECTION_DISPLAY) ;

  return 1 ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		command_disconnectUnits
 *	Description:	disconnects two units
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_disconnectUnit - 0 on failure, 1 on success
 ***********************************************************************/
int	command_disconnectUnits (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  Unit	*preUnitArray,  *preUnit ;
  Unit	*postUnitArray, *postUnit ;

  IUsage("<unit1> <unit2>") ;
  if (GiveHelp(tokc)) {
    ISynopsis("disconnects two sets of units from each other") ;
    IHelp
      (IHelpArgs,
       "Disconnects <unit1> from <unit2>,  where <unit1> and <unit2> are the",
       "names  of two  units in  the  current network.   Order is important.",
       "<unit1>  is  the  preceding unit  to  <unit2>.  Unit  names may  use",
       "regular expressions  similar to  grep  to  disconnect  more than two",
       "units.",
       "EXAMPLES",
       "\txerion-> disconnectUnits \"bias\" \"hidden.*\"",
       "",
       "will disconnect the unit with name  \"bias\" from all units whose name",
       "begins with the string \"hidden\".",
       "SEE ALSO",
       "connectGroups, connectUnits, disconnectGroups",
       NULL);
    return 0;
  }

  if (tokc != 3) {
    IErrorAbort(IPrintUsage(tokv[0], usage)) ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  /* get the list of pre groups assuming tokv[1] is a regular expression */
  preUnitArray = unitsFromRegex(currentNet, tokv[1]) ;

  if (preUnitArray == NULL)
    IErrorAbort("Bad string: \"%s\"", tokv[1]) ;
  if (preUnitArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", tokv[1]) ;

  /* get the list of post units assuming tokv[2] is a regular expression */
  postUnitArray = unitsFromRegex(currentNet, tokv[2]) ;
  if (postUnitArray == NULL)
    IErrorAbort("Bad string: \"%s\"", tokv[2]) ;
  if (postUnitArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", tokv[2]) ;

  for (preUnit = preUnitArray ; *preUnit != NULL ; ++preUnit) {
    for (postUnit = postUnitArray ; *postUnit != NULL ; ++postUnit) {
      disconnectUnits(*preUnit, *postUnit) ;
    }
  }

  free(preUnitArray) ;
  free(postUnitArray) ;

  markToRebuildDisplay(CONNECTION_DISPLAY) ;

  return 1 ;
}
/**********************************************************************/
