
/**********************************************************************
 * $Id: varsub.c,v 1.4 92/04/20 16:10:59 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 **********************************************************************/

#include "itf.h"
#include "varsub.h"

/* #define DBP(X) printf X */
#define DBP(X)

/*
 * VarSubstitute()
 *
 * Do variable subsitution on a line.  Neither quotes nor backslashes are
 * removed from the line.  The line is terminated by null.
 *
 * A variable is identified by the $ sign in front of it
 * Quotes are noticed, variables are substituted inside double quotes but
 * not inside single quotes.
 * A variable may be written as $var or ${var}. Variables must be composed
 * solely of alphanumeric characters and (including _).
 */

static int source_level = 0;

char *VarSubstitute(line,output,outlen)
char *line;
char *output;
int outlen;
{
  int c;
  int in_single_quotes=0;
  int in_double_quotes=0;
  int in_var_name=0;
  int look_for_brace=0;
  int escaped;
  int escape_pending=0;
  char *outend = output+outlen;
  char *outstart = output;
  char *value;
  char *var_char;	/* current position in the var name */
  static char var_name[MAX_VAR_NAME];
  var_char = 0;
  if (index(line,'$')==NULL) return line;	/* most times return here */
  while ((c = *line) || in_var_name) {
    if (*line) line++;
    if (escape_pending) escaped=1; else escaped=0;
    escape_pending=0;
    if (in_var_name) {
      if ((look_for_brace&&c!='}'&&c!=0) || isalnum(c) || c=='_' || c=='-') {
	/*
	 * still in a var name
	 */
	*(var_char++) = c;
	if (var_char-var_name > MAX_VAR_NAME-1) {
	  *(--var_char) = '\0';
	  IError("Variable name too long: \"%s...\"",var_name);
	  return 0;
	}
      } else if (c=='{' && !look_for_brace && var_char==var_name) {
	look_for_brace=1;
      } else {
	/*
	 * at the end of a var name
	 * deal with the current character on the next iteration of the 'while'
	 */
	if (*line && c!='}') line--;
	in_var_name = 0;
	*var_char = 0;
	if ((look_for_brace && c!='}')
	    || (look_for_brace && var_char==var_name)) {
	  *var_char = '\0';
	  IError("Missing right '}' on: \"%s\"",var_name);
	  return NULL;
	}
	if (var_char==var_name) *(output++) = '$'; /* an isolated '$' sign */
	else {
	  int is_parameter = 0;
	  if (IIsInteger(var_name)) { /* a parameter - put in level */
	    sprintf(var_name,"@%d@%d@",source_level,atoi(var_name));
	    is_parameter = 1;
	  }
	  DBP(("Raw variable name is %s\n",var_name));
	  value = IGetValue(var_name,NULL);
	  if (itf_errno) {
	    if (is_parameter)
	      IResetError();	/* just skip this parameter - value is "" */
	    else
	      return NULL;
	  } else {
	    DBP(("Processed variable name is %s, value is %s\n",var_name,value));
	    if (output+strlen(value) >= outend) {
	      IError("Variable-substituted command is too long");
	      return NULL;
	    }
	    (void) strcpy(output,value);
	    output += strlen(value);
	  }
	}
      }
    } else if (!escaped) {
      switch (c) {
      case '\\':
	/* escape character - set escape pending */
	escape_pending = 1;
	*(output++) = c;
	break;
      case '\'':
	/* single quote character */
	if (in_single_quotes) in_single_quotes=0;
	else if (!in_double_quotes) in_single_quotes=1;
	*(output++) = c;
	break;
      case '"':
	/* double quote character */
	if (in_double_quotes) in_double_quotes=0;
	else if (!in_single_quotes) in_double_quotes=1;
	*(output++) = c;
	break;
      case '$':
	if (!in_single_quotes) {
	  /* start of variable */
	  in_var_name = 1;
	  var_char = var_name;
	  look_for_brace=0;
	} else
	  *(output++) = c;
	break;
      case '\0':
	if (in_single_quotes || in_double_quotes) {
	  /* unmatched something */
	  IError("Unmatched quote '%c'",
		 (in_single_quotes ? '\'' : (in_double_quotes ? '"' : '?')));
	  return NULL;
	}
	break;
      default:
      /* ordinary character in the line */
	*(output++) = c;
      }
    } else
	*(output++) = c;
    if (output >= outend) {
      IError("Variable substituted command too long");
      return NULL;
    }
  }
  *output = '\n';
  ++output ;
  *output = '\0';
  return outstart;
}
