
/**********************************************************************
 * $Id: connect.c,v 1.12 92/04/20 16:15:41 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <values.h>
#include <xerion/simulator.h>
#include "connect.h"

static void	intraconnectGroup	ARGS((Group group)) ;
static void	connectToGroup  	ARGS((Unit  unit, void	*data)) ;
static void	connectFromUnit 	ARGS((Unit  unit, void	*data)) ;
static void	disconnectFromPostGroup ARGS((Unit  unit, void	*data)) ;
static void	disconnectFromPreUnit   ARGS((Unit  unit, void	*data)) ;

/***********************************************************************
 *	Name:		connectUnits
 *	Description:	connects together to units by creating a link
 *			and setting all the appropriate fields
 *	Parameters:	
 *		Unit	preUnit  - the unit the link is to come from
 *		Unit	postUnit - the unit the link is to go to
 *		Mask	mask     - a mask to use when creating the link
 *	Return Value:	
 *		the link created
 ***********************************************************************/
Link	connectUnits (preUnit, postUnit, mask)
  Unit	preUnit ;
  Unit	postUnit ;
  Mask	mask ;
{
  char	name[128] ;
  
  sprintf(name, "%s %s %s", 
	  preUnit->name, 
	  (mask&BIDIRECTIONAL ? "<->" : "->"),
	  postUnit->name) ;
  return createLink (name, preUnit, postUnit, mask) ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		connectGroups
 *	Description:	connects two groups completely
 *	Parameters:	
 *		Group	preGroup   - the group the links are to come from
 *		Group	postGroup  - the group the links are to go to
 *		Mask	mask       - a mask to use when creating the links
 *		Real	prob	   - probability of creating links
 *	Return Value:	
 ***********************************************************************/
static Mask	linkMask ;
static Real	linkProb ;
static int	status ;
int	connectGroups (preGroup, postGroup, mask, prob)
  Group	preGroup ;
  Group	postGroup ;
  Mask	mask ;
  double prob ;
{
  linkMask = mask ;
  linkProb = prob ;	/* hack to pass global */
  status = 1 ;
  if (preGroup==postGroup && linkMask&BIDIRECTIONAL) 
    intraconnectGroup(postGroup) ;
  else
    groupForAllUnits(preGroup, connectToGroup, postGroup) ;
  return status ;
}
/***********************************************************************
 *	Name:		topoConnectGroups
 *	Description:	connects two groups completely
 *	Parameters:	
 *		Group	preGroup   - the group the links are to come from
 *		Group	postGroup  - the group the links are to go to
 *		Mask	mask       - a mask to use when creating the links
 *		Real	prob	   - probability of creating links
 *	Return Value:	
 ***********************************************************************/
int	topoConnectGroups (preGroup, postGroup, mask, prob,
			   preStart, preNum, postStart, postNum)
  Group	preGroup ;
  Group	postGroup ;
  Mask	mask ;
  double prob ;
  int	preStart, preNum, postStart, postNum ;
{
  int	preIdx, postIdx ;
  Unit	*preUnit, *postUnit ;

  linkMask = mask ;
  linkProb = prob ;	/* hack to pass global */
  status = 1 ;
  if (!preNum)  preNum  = preGroup->numUnits  - preStart ;
  if (!postNum) postNum = postGroup->numUnits - postStart ;

  for (preIdx = 0, preUnit = &preGroup->unit[preStart] ;
       preIdx < preNum ;
       preIdx++, preUnit++)
    for (postIdx = 0, postUnit = &postGroup->unit[postStart] ;
	 postIdx < postNum ;
	 postIdx++, postUnit++)
      if ((preUnit != postUnit || linkMask&ALLOWSELF) && 
	  WITH_PROB(linkProb)	&&
	  connectUnits(*preUnit, *postUnit, linkMask) == NULL)
	status = 0 ;
  return status ;
}
/**********************************************************************/
static void	connectToGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	postGroup = (Group)data ;
  
  groupForAllUnits(postGroup, connectFromUnit, unit) ;
}
/**********************************************************************/
static void	connectFromUnit(unit, data)
  Unit		unit ;
  void		*data ;
{
  Unit	preUnit = (Unit)data ;
  if ((preUnit != unit || linkMask&ALLOWSELF) &&
      WITH_PROB(linkProb)	&&
      connectUnits(preUnit, unit, linkMask) == NULL)
    status = 0 ;
}
/**********************************************************************/
/***********************************************************************
 *	Name:		intraconnectGroup
 *	Description:	Connects a group with itself
 *			Currently this will only be called if 
 *			linkMask&BIDIRECTIONAL
 *	Parameters:	
 *		Group	group   - the group to intraconnect
 *	Return Value:	NONE
 ***********************************************************************/
static void	intraconnectGroup (group)
  Group	group ;
{
  int	preIdx, postIdx ;
  Unit	*preUnit, *postUnit ;
  
  for (postIdx = 0, postUnit = group->unit ;
       postIdx < group->numUnits ; ++postIdx, ++postUnit)
    for (preIdx = 0, preUnit = group->unit ;
	 preIdx < postIdx ; ++preIdx, ++preUnit)
      if (WITH_PROB(linkProb)	&&
	  connectUnits(*preUnit, *postUnit, linkMask) == NULL)
	status = 0 ;
  if (linkMask&ALLOWSELF) 
    for (postIdx = 0, postUnit = group->unit ;
	 postIdx < group->numUnits ; ++postIdx, ++postUnit) 
      if (WITH_PROB(linkProb)	&&
	  connectUnits(*postUnit, *postUnit, linkMask) == NULL)
	status = 0 ;
}

/***********************************************************************
 *	Name:		disconnectUnits
 *	Description:	disconnects two units by destroying the
 *			link between them.
 *	Parameters:	
 *		Unit	preUnit  - the unit the link is to come from
 *		Unit	postUnit - the unit the link is to go to
 *	Return Value:	
 *		NONE
 ***********************************************************************/
int	disconnectUnits (preUnit, postUnit)
  Unit	preUnit ;
  Unit	postUnit ;
{
  int	idx ;
  
  for (idx = 0 ; idx < preUnit->numOutgoing ; ++idx) {
    if (postUnit == preUnit->outgoingLink[idx]->postUnit) {
      destroyLink(preUnit->outgoingLink[idx]) ;
      break ;
    }
  }
  return 1 ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		disconnectGroups
 *	Description:	disconnects two groups completely
 *	Parameters:	
 *		Group	preGroup   - the group the links are to come from
 *		Group	postGroup  - the group the links are to go to
 *	Return Value:	
 ***********************************************************************/
int	disconnectGroups (preGroup, postGroup)
  Group	preGroup ;
  Group	postGroup ;
{
  groupForAllUnits(preGroup, disconnectFromPostGroup, postGroup) ;
  return 1 ;
}
/**********************************************************************/
static void	disconnectFromPostGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group		postGroup = (Group)data ;
  
  groupForAllUnits(postGroup, disconnectFromPreUnit, unit) ;
}
/**********************************************************************/
static void	disconnectFromPreUnit(unit, data)
  Unit		unit ;
  void		*data ;
{
  Unit		preUnit = (Unit)data ;
  disconnectUnits(preUnit, unit) ;
}
/**********************************************************************/
