
/**********************************************************************
 * $Id: example.c,v 1.12 92/04/20 16:15:46 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <strings.h>

#include <xerion/simulator.h>
#include "simUtils.h"
#include "example.h"

/***********************************************************************
 *	Private procedures
 ***********************************************************************/
static void	initialize ARGS((ExampleSet	exampleSet)) ;
static void	finalize   ARGS((ExampleSet	exampleSet)) ;

static int	reset      ARGS((ExampleSet	exampleSet)) ;
static int	getNext    ARGS((ExampleSet	exampleSet)) ;
static int	getNumber  ARGS((ExampleSet	exampleSet, int	number)) ;
static int	addExampleList ARGS((ExampleSet exampleSet, int listSize, 
				     Real **input, int numInputs, 
				     Real **target, int numTargets)) ;

static void	setInput   ARGS((Unit	unit,  void	*data)) ;
static void	setTarget  ARGS((Unit	unit,  void	*data)) ;
static void	sumUnits   ARGS((Group	group, void	*data)) ;
/**********************************************************************/


/***********************************************************************
 *	Private variables
 ***********************************************************************/
static ExampleSetProc	createExampleSetHook  = NULL ;
static ExampleSetProc	destroyExampleSetHook = NULL ;
/**********************************************************************/


/***********************************************************************
 *	Name:		setCreateExampleSetHook
 *	Description:	sets the user hook to be called whenever an exampleSet
 *			is created.
 *	Parameters:	
 *		ExampleSetProc	hook - the procedure to be called:
 *				void hook  (ExampleSet	exampleSet)
 *	Return Value:	
 *		NONE
 ***********************************************************************/
void	setCreateExampleSetHook(hook)
  ExampleSetProc	hook ;
{
  createExampleSetHook = hook ;
}
/**********************************************************************/
void	setDestroyExampleSetHook(hook)
  ExampleSetProc	hook ;
{
  destroyExampleSetHook = hook ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		createExampleSet 
 *	Description:	creates an exampleSet and initializes all of its data
 *			as well as inserting it into a network
 *	Parameters:	
 *		char	*name - the name for the exampleSet
 *		Net	net   - the net to insert the exampleSet in
 *	Return Value:	
 *		ExampleSet	createExampleSet - the new exampleSet
 ***********************************************************************/
ExampleSet	createExampleSet (name, mask, net)
  String	name ;
  Mask		mask ;
  Net		net ;
{
  ExampleSet	exampleSet ;

  exampleSet = (ExampleSet)malloc (sizeof(ExampleSetRec)) ;

  if (exampleSet == NULL) {
    simError(SIM_ERR_MALLOC, "createExampleSet \"%s\"", name) ;
    return exampleSet ;
  }

  exampleSet->name = strdup (name) ;
  exampleSet->type = mask ;
  exampleSet->net  = net ;

  initialize (exampleSet) ;

  if (createExampleSetHook != NULL)
    (*createExampleSetHook) (exampleSet) ;

  return exampleSet ;
}
/***********************************************************************
 *	Name:		initialize
 *	Description:	init proc for the ExampleSet class of objects
 *		ExampleSet	exampleSet - the exampleSet object ;
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void	initialize (exampleSet) 
  ExampleSet	exampleSet ;
{
  exampleSet->resetAction     = reset ;
  exampleSet->getNextAction   = getNext ;
  exampleSet->getNumberAction = getNumber ;
  exampleSet->addExampleList  = addExampleList ;
 
  exampleSet->example          = NULL ;
  exampleSet->numExamples      = 0 ;
  exampleSet->currentIdx       = -1 ;
  exampleSet->exampleArraySize = 0 ;

  exampleSet->clampMask  = INPUT ;
  exampleSet->targetMask = OUTPUT ;

  exampleSet->extension = NULL ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		destroyExampleSet 
 *	Description:	destroys a exampleSet
 *	Parameters:	
 *		ExampleSet	exampleSet - the exampleSet to destroy
 *	Return Value:	
 *		NONE
 ***********************************************************************/
void	destroyExampleSet (exampleSet) 
  ExampleSet	exampleSet ;
{
  if (destroyExampleSetHook != NULL)
    (*destroyExampleSetHook) (exampleSet) ;

  finalize (exampleSet) ;

  free (exampleSet->name) ;
  free (exampleSet) ;
}
/***********************************************************************
 *	Name:		finalize
 *	Description:	finalize proc for the exampleSet class of objects
 *		ExampleSet	exampleSet - the exampleSet object ;
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void	finalize (exampleSet) 
  ExampleSet	exampleSet ;
{
  if (exampleSet->example != NULL) {
    Example	example = exampleSet->example ;
    int		idx ;

    /* for all the examples */
    for (idx = 0 ; idx < exampleSet->numExamples ; ++idx) {
      int	vectorIdx ;
      int	nextVectorIdx ;
      if (example[idx].inputVector != NULL) {
	int	numVectors = example[idx].numVectors ;

	/* for all the input vectors in an example */
	for (vectorIdx = 0 ; vectorIdx < numVectors ; ++vectorIdx) {
	  Real	*vectorPtr = example[idx].inputVector[vectorIdx] ;

	  if (vectorPtr != NULL) {
	    free(vectorPtr) ;

	    /* if the vector was shared, set all other pointers to NULL */
	    for (nextVectorIdx = vectorIdx + 1 ; 
		 nextVectorIdx < numVectors ; ++nextVectorIdx) {
	      if (vectorPtr == example[idx].inputVector[nextVectorIdx])
		example[idx].inputVector[nextVectorIdx] = NULL ;
	    }
	  }
	}
	free(example[idx].inputVector) ;
      }

      if (example[idx].targetVector != NULL) {
	int	numVectors = example[idx].numVectors ;

	/* for all the target vectors in an example */
	for (vectorIdx = 0 ; vectorIdx < numVectors ; ++vectorIdx) {
	  Real	*vectorPtr = example[idx].targetVector[vectorIdx] ;

	  if (vectorPtr != NULL) {
	    free(vectorPtr) ;

	    /* if the vector was shared, set all other pointers to NULL */
	    for (nextVectorIdx = vectorIdx + 1 ; 
		 nextVectorIdx < numVectors ; ++nextVectorIdx) {
	      if (vectorPtr == example[idx].targetVector[nextVectorIdx])
		example[idx].targetVector[nextVectorIdx] = NULL ;
	    }
	  }
	}
	free(example[idx].targetVector) ;
      }
    free (exampleSet->example) ;
    }
  }
}
/**********************************************************************/


/***********************************************************************
 *	Name:		exampleSetAddExample 
 *	Description:	adds an example to an example set.
 *	Parameters:	
 *		ExampleSet	exampleSet - the example set to add to
 *		Real		*input - the array of inputs to add
 *		int		numInputs - the number of input values
 *		Real		*target - the array of targets to add
 *		int		numTargets - the number of target values
 *	Return Value:	
 *		NONE
 ***********************************************************************/
int	exampleSetAddExample (exampleSet, input, numInputs, target, numTargets)
  ExampleSet	exampleSet ;
  Real		*input ;
  int		numInputs ;
  Real		*target ;
  int		numTargets ;
{
  Real		**inputVector, **targetVector ;

  inputVector	 = (Real **)malloc(sizeof(Real *)) ;
  inputVector[0] = input ;

  targetVector    = (Real **)malloc(sizeof(Real *)) ;
  targetVector[0] = target ;

  return exampleSet->addExampleList(exampleSet, 1, 
				    inputVector, numInputs, 
				    targetVector, numTargets) ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		exampleSetAddExampleList
 *	Description:	
 *	Parameters:
 *	  ExampleSet	exampleSet - 
 *	  int		listSize - 
 *	  Real		**input - 
 *	  int		numInputs - 
 *	  Real		**target - 
 *	  int		numTargets - 
 *	Return Value:
 *	  int	  exampleSetAddExampleList - 
 *********************************************************************/
int	  exampleSetAddExampleList(exampleSet, listSize,  input,  
				   numInputs, target, numTargets)
  ExampleSet	exampleSet ;
  int		listSize ;
  Real		**input ;
  int		numInputs ;
  Real		**target ;
  int		numTargets ;
{
  return exampleSet->addExampleList(exampleSet, listSize, 
				    input, numInputs, 
				    target, numTargets) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		addExampleList
 *	Description:	Method for adding examples to the set. NOTE:
 *			the example arrays are used in place. They must
 *			be allocated using malloc, and remain in existence
 *			as long as the example set does. (it will free them
 *			when it is done).
 *	Parameters:
 *	  ExampleSet	exampleSet - the example set
 *	  int		listSize - the size of the list (1 for normal
 *				   nets, larger for recurrent nets
 *	  Real		**input  - the array (dimension listSize) of
 *				   input vectors.
 *	  int		numInputs - the dimension of each input vector
 *	  Real		**target - the array (dimension listSize) of
 *				   target vectors.
 *	  int		numTargets - the dimension of each target vector
 *	Return Value:
 *	  static int	addExampleList - 1
 *********************************************************************/
#ifndef GRANULARITY
#define GRANULARITY 64
#endif
static int	addExampleList(exampleSet, listSize,  input,  
			       numInputs, target, numTargets)
  ExampleSet	exampleSet ;
  int		listSize ;
  Real		**input ;
  int		numInputs ;
  Real		**target ;
  int		numTargets ;
{
  Example	example ;

  /* Make sure that the arrays are big enough */
  if (exampleSet->numExamples >= exampleSet->exampleArraySize) {
    if (exampleSet->exampleArraySize == 0) {
      /* set array size to some default size */
      exampleSet->exampleArraySize = exampleSet->numExamples + GRANULARITY ;
      exampleSet->example 
	= (ExampleRec *)malloc ((exampleSet->exampleArraySize)
				*sizeof(ExampleRec)) ;
    } else {
      /* double the size of the array */
      exampleSet->exampleArraySize = 2.0*exampleSet->numExamples ;
      exampleSet->example 
	= (ExampleRec *)realloc (exampleSet->example,
				 (exampleSet->exampleArraySize)
				 *sizeof(ExampleRec)) ;
    }
    if (exampleSet->example == NULL) {
      simError(SIM_ERR_MALLOC, "addExampleList \"%s\"",
	       exampleSet->name) ;
      return 0 ;
    }
  }

  example = &exampleSet->example[exampleSet->numExamples] ;

  example->inputVector = input ;
  example->numInputs   = numInputs ;

  example->targetVector = target ;
  example->numTargets   = numTargets ;

  example->numVectors = listSize ;

  example->name		= NULL ;
  example->frequency	= 1.0 ;

  ++(exampleSet->numExamples) ;
  return 1 ;
}
/********************************************************************/
/***********************************************************************
 *	Name:		reset
 *	Description:	resets the example list in an example set
 *	Parameters:	
 *		ExampleSet	exampleSet - the exampleSet
 *	Return Value:	
 *		int		reset - the index of the current example
 ***********************************************************************/
static int	reset (exampleSet) 
  ExampleSet	exampleSet ;
{
  exampleSet->currentIdx = -1 ;
  return exampleSet->currentIdx ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		getNext
 *	Description:	getNext action for the exampleSet class of objects 
 *			objects. It sets the network values and also
 *			sets the current example number.p
 *		ExampleSet	exampleSet - the exampleSet object ;
 *	Return Value:	
 *		int	getNext - the index of this example, -1 if none
 *				left
 ***********************************************************************/
static int	unitIdx ;
/**********************************************************************/
static int	getNext (exampleSet) 
  ExampleSet	exampleSet ;
{
  int		numClamps, numTargets ;
  Example	example ;
  Mask		clampMask, targetMask ;

  if (exampleSet->numExamples == 0)
    return -1 ;

  ++exampleSet->currentIdx ;
  if (exampleSet->currentIdx == exampleSet->numExamples)
    exampleSet->currentIdx = 0 ;

  example = &exampleSet->example[exampleSet->currentIdx] ;

  /* Make sure that the number of time slices is the same
   * in the example and the net */
  if (example->numVectors != exampleSet->net->timeSlices) {
    fprintf(stderr, 
	    "Inconsistent number of time slices (%d:%d) in net \"%s\"\n", 
	    example->numVectors, exampleSet->net->timeSlices, 
	    exampleSet->net->name) ;
    return exampleSet->currentIdx ;
  }

  /* make sure that the clamped units exist and there is 
   * the same number as the exampleSet's input array */
  clampMask = exampleSet->clampMask ;
  numClamps = 0 ;
  netForAllGroups(exampleSet->net, clampMask, sumUnits, &numClamps) ;

  if (numClamps == 0) {
    fprintf(stderr, "No clamped units in net \"%s\"\n", 
	    exampleSet->net->name) ;
    return exampleSet->currentIdx ;
  }
  if (numClamps != example->numInputs) {
    fprintf(stderr, "Inconsistent number of clamped units in net \"%s\"\n", 
	    exampleSet->net->name) ;
    return exampleSet->currentIdx ;
  }

  unitIdx = 0 ;
  netForAllUnits(exampleSet->net, clampMask, setInput, example->inputVector);

  /* only do targets if they exist */
  if (example->numTargets == 0)
    return exampleSet->currentIdx ;

  /* make sure that the target units exist and there is 
   * the same number as the example's target array */
  targetMask = exampleSet->targetMask ;
  numTargets = 0 ;
  netForAllGroups(exampleSet->net, targetMask, sumUnits, &numTargets) ;
  if (numTargets == 0) {
    fprintf(stderr, "No target units in net \"%s\"\n", exampleSet->net->name) ;
    return exampleSet->currentIdx ;
  }
  if (numTargets != example->numTargets) {
    fprintf(stderr, "Inconsistent number of target units in net \"%s\"\n", 
	    exampleSet->net->name) ;
    return exampleSet->currentIdx ;
  }

  unitIdx = 0 ;
  netForAllUnits(exampleSet->net,targetMask,setTarget, example->targetVector) ;

  return exampleSet->currentIdx ;
}
/**********************************************************************/
static void	setInput (unit, data)
  Unit		unit ;
  void		*data ;
{
  if (unit->net->type & RECURRENT) {
    int		idx, maxIdx = unit->timeSlices ;
    Real	**vector = ((Real **)data) ;
    for (idx = 0 ; idx < maxIdx ; ++idx)
      unit->clampingHistory[idx] = vector[idx][unitIdx] ;
  } else {
    unit->output = ((Real **)data)[0][unitIdx] ;
  }
  ++unitIdx ;
}
/**********************************************************************/
static void	setTarget (unit, data)
  Unit		unit ;
  void		*data ;
{
  if (unit->net->type & RECURRENT) {
    int		idx, maxIdx = unit->timeSlices ;
    Real	**vector = ((Real **)data) ;
    for (idx = 0 ; idx < maxIdx ; ++idx)
      unit->targetHistory[idx] = vector[idx][unitIdx] ;
  } else {
    unit->target = ((Real **)data)[0][unitIdx] ;
  }
  ++unitIdx ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		getNumber
 *	Description:	getNumber action for the exampleSet class of 
 *			objects. It sets the network values and also
 *			sets the current example number.p
 *		ExampleSet	exampleSet - the exampleSet object ;
 *		int		number     - the index (base 0) of the
 *					required example.
 *	Return Value:	
 *		int	getNumber - the index of this example, -1 if
 *				it doesn't exist
 ***********************************************************************/
static int	getNumber (exampleSet, number)
  ExampleSet	exampleSet ;
  int		number ;
{
  if (number >= 0 && number < exampleSet->numExamples) {
    exampleSet->currentIdx = number - 1 ;
    return MgetNext(exampleSet) ;
  } else {
    return -1 ;
  }
}
/**********************************************************************/


/***********************************************************************
 *	Name:		sumUnits
 *	Description:	increments an integer (address passed in as data)
 *			by the number of units in the group
 *	Parameters:	
 *		Group	group - the group to sum
 *		void	*data - a pointer to an int to increment.
 *	Return Value:	NONE
 ***********************************************************************/
static void	sumUnits (group, data)
  Group		group ;
  void		*data ;
{
  *(int *)data += group->numUnits ;
}
/**********************************************************************/
