/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <math.h>
#include <xerion/itf.h>
#include <xerion/minimize.h>
#include <commands.h>

int command_checkpoint(tokc,tokv)
int tokc;
char *tokv[];
{
  char	command[100] ;
  int	epoch ;
  double tmp ;

  IUsage("<filename root>");
  if (GiveHelp(tokc)) {
    ISynopsis("Checkpoint weights of network into a file");
    IHelp
      (IHelpArgs,
       "Save the weights of the current network into a specified file.",
       "The actual file name will appended with the current epoch number",
       "and the extension \".wts\".  This is best used as a trace.",
       "EXAMPLE",
       "\txerion-> checkpoint encoder",
       NULL);
    return 0;
  }
  if (tokc != 2) IErrorAbort(IPrintUsage(tokv[0], usage)) ;
  if (currentNet == NULL) IErrorAbort("No current network") ;

  epoch = currentNet->mz->nIterations ;

  if (epoch >= 10 && epoch%((int)pow(10,floor(log10((double)epoch))))==0) {
    fprintf(dout,"Checkpointing file \"%s.%d.wts\".\n", tokv[1], epoch) ;
    sprintf(command, "sh rm -f %s.%d.wts; saveWeights > %s.%d.wts", tokv[1], epoch, tokv[1], epoch) ;
    IDoCommandLine(command) ;
  }
  return 1 ;
}
/****************************************************************************
 * PDP weight files (savePDPweights, loadPDPweights)
 ****************************************************************************/
static	void	readWt(link, data)
  Link	link ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  double wt ;

  if (! (link->preUnit->group->type & BIAS)) {
    if (fscanf(fp, "%lf", &wt)==EOF) {
      fclose(fp) ;
      IErrorAbort("Error reading PDP weight for link %s", link->name) ;
    } else 
      link->weight = (Real)wt ;
  }
}
static	void	writeWt(link, data)
  Link	link ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;

  if (! (link->preUnit->group->type & BIAS))
    fprintf(fp, "%f\n", link->weight) ;
}

static	void	readWeights(unit, data)
  Unit	unit ;
  void	*data ;
{
  unitForAllLinks(unit, INCOMING, readWt, data) ;
}

static	void	writeWeights(unit, data)
  Unit	unit ;
  void	*data ;
{
  unitForAllLinks(unit, INCOMING, writeWt, data) ;
}

static	void	readBias(unit, data)
  Unit	unit ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  double wt ;

  if (fscanf(fp, "%lf", &wt)==EOF) 
    IErrorAbort("Error reading PDP bias for unit %s", unit->name) ;
  else if (unit->numIncoming &&
	   unit->incomingLink[0]->preUnit->group->type & BIAS) {
    /* units w/o biases (i.e. input units) will have 0.0 in file */
    unit->incomingLink[0]->weight = (Real)wt ;
  }
}

static	void	writeBias(unit, data)
  Unit	unit ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  double wt ;

  if (unit->numIncoming &&
      unit->incomingLink[0]->preUnit->group->type & BIAS)
    fprintf(fp, "%f\n", unit->incomingLink[0]->weight) ;
  else
    fprintf(fp, "0.0\n") ;
}
/****************************************************************************/
int command_loadPDPweights (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  FILE		*fp ;
  String	filename = NULL ;
  int		numUnits = 0 ;
  struct ARG_DESC *argd ;
  static int	*unitNum, staticNumUnits = 0 ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[<file-name>%s]",	   &filename, 
       "file to read from (defaults to stdin)");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("read weights from PDP weight file (or stdin).") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  if (!currentNet) IErrorAbort("No current network") ;

  if (filename) fp = IOpenFileOrAbort(filename, "r", NULL) ;
  else          fp = din ;
  
  netForAllOtherUnits(currentNet, BIAS, readWeights, (void *)fp) ;
  netForAllOtherUnits(currentNet, BIAS, readBias,    (void *)fp) ;
  markToRebuildDisplay(CONNECTION_DISPLAY) ;

  if (filename) fclose(fp) ;
  return 1 ;
}
/****************************************************************************/
int command_savePDPweights (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  FILE		*fp ;
  String	filename = NULL ;
  int		numUnits = 0 ;
  struct ARG_DESC *argd ;
  static int	*unitNum, staticNumUnits = 0 ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[<file-name>%s]",	   &filename, 
       "file to write to (defaults to stdout)");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("write weights to PDP weight file (or stdout).") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  if (!currentNet) IErrorAbort("No current network") ;

  if (filename) fp = IOpenFileOrAbort(filename, "w", NULL) ;
  else          fp = dout ;
  
  netForAllOtherUnits(currentNet, BIAS, writeWeights, (void *)fp) ;
  netForAllOtherUnits(currentNet, BIAS, writeBias,    (void *)fp) ;

  if (filename) fclose(fp) ;
  return 1 ;
}
/****************************************************************************/
static	void	readState(unit, data)
  Unit  unit ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  double st ;

  if (fscanf(fp, "%lf", &st)==EOF) {
    fclose(fp) ;
    IErrorAbort("Error reading state for unit %s", unit->name) ;
  } else 
    unit->output = (Real)st ;
}
static	void	readStates(group, data)
  Group group ;
  void	*data ;
{
  groupForAllUnits(group, readState, data) ;
}
static	void	writeState(unit, data)
  Unit	unit ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  fprintf(fp, "%f ", unit->output) ;
}
static	void	writeStates(group, data)
  Group group ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  groupForAllUnits(group, writeState, data) ;
  fprintf(fp, "\n") ;
}
/****************************************************************************/
int	command_saveStates (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  FILE		*fp ;
  String	filename = NULL ;
  String groupRE = NULL ;
  Group	*groupArray, *group ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[-Pgroup <regexp>%s%?]",	&groupRE,
       "regular expression specifying destination groups.");
  Args(argd, "[<file-name>%s]",	   &filename, 
       "file to read from (defaults to stdin)");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("save states of some units") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
      NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (filename) fp = IOpenFileOrAbort(filename, "w", NULL) ;
  else          fp = dout ;

  if (!groupRE) groupRE = ".*" ;
  groupArray = groupsFromRegex(currentNet, groupRE) ;
  if (groupArray == NULL) IErrorAbort("Bad string: \"%s\"", groupRE) ;
  if (groupArray[0] == NULL) IErrorAbort("No match: \"%s\"", groupRE) ;
  for (group = groupArray ; *group != NULL ; ++group) {
    fprintf(fp, "%s\n", (*group)->name) ;
    writeStates(*group, (void *)fp) ;
  }

  if (filename) fclose(fp) ;
  free(groupArray) ;
  return 1 ;
}
/**********************************************************************/
int command_loadStates (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  FILE		*fp ;
  String	filename = NULL ;
  char		groupname[128] ;
  Group		group ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[<file-name>%s]",	   &filename, 
       "file to read from (defaults to stdin)");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("load states of units") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
      NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);

  if (filename) fp = IOpenFileOrAbort(filename, "r", NULL) ;
  else          fp = din ;
  
  while (fscanf(fp, "%s", groupname) != EOF) {
    if ((group=groupFromName(currentNet, groupname)) == NULL)
      IErrorAbort("unknown group \"%s\".", groupname) ;
    readStates(group, (void *)fp) ;
  }
  markToRebuildDisplay(ACTIVITY_DISPLAY) ;

  if (filename) fclose(fp) ;
  return 1 ;
}
/****************************************************************************/
