#include <useful.h>
#include <simulator.h>
#include <commands.h>
#include <values.h>
#include <nan.h>
#include "utilsDCP.h"

extern void	lesionUnit	ARGS((Unit  unit)) ;
extern void	unlesionUnit	ARGS((Unit  unit)) ;
extern Boolean	isLesionedUnit	ARGS((Unit  unit)) ;
extern void	lesionLink	ARGS((Link  link)) ;
extern void	unlesionLink	ARGS((Link  link)) ;
extern Boolean	isLesionedLink	ARGS((Link  link)) ;

static	Real	probability ;
static	Real	minWeightMag ;

#define MoldWeight(link)	((link)->deltaWeight)

/***********************************************************************/
static void	corruptWeightsFromGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	preGroup = (Group)data ;
  int	idx ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link        link    = incoming[idx] ;
    if (link->preUnit->group == preGroup) {
      MoldWeight(link) = link->weight ;
      link->weight += gaussianNoise(probability) ;
    }
  }
}
/***********************************************************************/
static void	restoreWeightsFromGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	preGroup = (Group)data ;
  int	idx ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link        link    = incoming[idx] ;
    if (link->preUnit->group == preGroup) 
      link->weight = MoldWeight(link) ;
  }
}
/***********************************************************************/
void	corruptWeights (preGroup, postGroup, sd)
  Group	preGroup ;
  Group	postGroup ;
  double sd ;
{
  probability = sd ;
  fprintf(dout,"Corrupting %s ==> %s weights with N(0,%.2g).\n",
	  preGroup->name, postGroup->name, sd) ;
  groupForAllUnits(postGroup, corruptWeightsFromGroup, preGroup) ;
  updateWeightDisplay() ;
}
/***********************************************************************/
void	restoreWeights (preGroup, postGroup)
  Group	preGroup ;
  Group	postGroup ;
{
  fprintf(dout,"Restoring %s ==> %s weights.\n", 
	  preGroup->name, postGroup->name) ;
  groupForAllUnits(postGroup, restoreWeightsFromGroup, preGroup) ;
  updateWeightDisplay() ;
}
/***********************************************************************
 *	Name:		lesionLinksFromGroup
 *	Description:	lesion links to unit from group
 *	Parameters:	
 *		Unit	unit - the unit receiving the links
 *		void	*data - the group sending the links
 *	Return Value:	NONE
 ***********************************************************************/
static void	lesionLinksFromGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	preGroup = (Group)data ;
  int	idx ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link        link    = incoming[idx] ;
    if (link->preUnit->group == preGroup) {
      if (isLesionedLink(link)) unlesionLink(link) ;
      if (WITH_PROBABILITY(probability) &&
	  fabs(link->weight) < minWeightMag)
	lesionLink(link) ;
    }
  }
}
static void	unitLesion(unit, data)
  Unit		unit ;
  void		*data ;
{
  if (WITH_PROBABILITY(probability))
    lesionUnit(unit) ;
  else
    unlesionUnit(unit) ;
}

/***********************************************************************
 *	Name:		lesion
 *	Description:	lesion links between two groups
 *	Parameters:	
 *		Group	preGroup   - the group the links are to come from
 *		Group	postGroup  - the group the links are to go to
 *		Real	prob       - proportion of links to lesion
 *	Return Value:	
 ***********************************************************************/
void	lesion (preGroup, postGroup, prob, minWt, print)
  Group	preGroup ;
  Group	postGroup ;
  double prob, minWt ;
  int 	print ;
{
  probability = prob ;
  minWeightMag = minWt ;
  if (postGroup) {	/* LINKS */
    if (print) {
      if (probability == 0.0) fprintf(dout,"Restoring ") ;
      else                    fprintf(dout,"Lesioning %.2g of ", probability) ;
      fprintf(dout,"%s ==> %s links", preGroup->name, postGroup->name) ;
      if (minWeightMag < ((Real) HUGE)) 
	fprintf(dout, " with |w|<%.4g", minWeightMag) ;
      fprintf(dout, "\n") ;
    }
    groupForAllUnits(postGroup, lesionLinksFromGroup, preGroup) ;
    updateWeightDisplay() ;
  } else {		/* UNITS */
    if (print) {
      if (probability == 0.0) fprintf(dout,"Restoring ") ;
      else                    fprintf(dout,"Lesioning %.2g of ", probability) ;
      fprintf(dout,"%s units\n", preGroup->name) ;
    }
    groupForAllUnits (preGroup, unitLesion, NULL) ;
  }
}

/***********************************************************************
 *	Name:		command_lesion
 *	Description:	(un)lesions units/connections
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	0 on failure, 1 on success
 ***********************************************************************/
int	command_lesion (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  Boolean lesioningLinks = FALSE ;
  int	noise = FALSE ;
  Real	sd = 0.0 ;
  String preGroupRE = NULL, postGroupRE = NULL ;
  Group	*preGroupArray,  *preGroup ;
  Group	*postGroupArray, *postGroup ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  probability = 1.0 ;
  minWeightMag = ((Real) HUGE) ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Pprobability <real>%r%?]",	&probability,	"probability of removing each specified link.");
  Args(argd, "[-PminWeightMag <real>%r%?]",	&minWeightMag,  "only lesion links with weight magnitudes below this.");
  Args(argd, "[-Pnoise %P%?]", 	    		&noise, 	"add noise to weights.");
  Args(argd, "[-Psd <real>%r%?]",		&sd,       	"standard deviation of noise added to weights.");
  Args(argd, "[-PpreGroup <regexp>%s%?]",	&preGroupRE,	"regular expression specifying source groups.");
  Args(argd, "[-PpostGroup <regexp>%s%?]",	&postGroupRE,   "regular expression specifying destination groups.");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("lesion some units or links between groups") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "Randomly lesion some proportion P of units in <preGroup>, or links",
       "between <preGroup> and <postGroup>, in the current network.  Group", 
       "names may  use regular expressions similar to those used in grep to ",
       "lesion more than one group or set of links." ,
       "Using P = 1.0 (the default) lesions all specified units/links;" ,
       "P = 0.0 restores all such units/links.", 
       "EXAMPLE",
       "\txerion-> lesion \"input\" \"hidden\"",
       "",
       "will lesion all of the links between the group \"input\" and the",
       "group \"hidden\".",
       "",
       "\txerion-> lesion -p 0.3 \"hidden*\"",
       "",
       "will lesion 0.3 of the units in all groups whose name begins with",
       "\"hidden\".",
      NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);

  preGroupArray = groupsFromRegex(currentNet, preGroupRE) ;
  if (preGroupArray == NULL)
    IErrorAbort("Bad string: \"%s\"", preGroupRE) ;
  if (preGroupArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", preGroupRE) ;

  if (preGroupRE != NULL) {
    lesioningLinks = TRUE ;
    postGroupArray = groupsFromRegex(currentNet, postGroupRE) ;
    if (postGroupArray == NULL)
      IErrorAbort("Bad string: \"%s\"", postGroupRE) ;
    if (postGroupArray[0] == NULL)
      IErrorAbort("No match: \"%s\"", postGroupRE) ;
  }
  for (preGroup = preGroupArray ; *preGroup != NULL ; ++preGroup) {
    if (lesioningLinks) 
      for (postGroup = postGroupArray ; *postGroup != NULL ; ++postGroup) {
	if (noise) {
	  if (sd == 0.0) restoreWeights(*preGroup, *postGroup) ;
	  else           corruptWeights(*preGroup, *postGroup, sd) ;
	} else
	  lesion(*preGroup, *postGroup, probability, minWeightMag, TRUE) ;
      } else {
	lesion(*preGroup, NULL, probability, minWeightMag, TRUE) ;
      }
  }
  free(preGroupArray) ;
  free(postGroupArray) ;
  
  return 1 ;
}
/**********************************************************************/
