/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <sys/times.h>
#include <math.h>
#include <values.h>
#include <nan.h>
#include <commands.h>
#include <xerion/minimize.h>
#include "utilsDCP.h"

Real	now()
{
  struct tms	tms;
  times(&tms) ;
  return (tms.tms_utime + tms.tms_stime)/100.0 ;
}

Real gaussianNoise(sd)
  double sd ;
{
  static int iset=FALSE;
  static Real gset;
  Real v1,v2,r,fac;

  if (sd == 0.0) return 0.0 ;
  if (iset) {
    iset = FALSE;
    return (gset*sd);
  } else {
    do {
      v1 = uniformNoise(-1.0,1.0);
      v2 = uniformNoise(-1.0,1.0);
      r  = v1*v1 + v2*v2;
    } while (r >= 1.0);
    fac = sqrt(-2.0*log(r)/r);
    gset = v1*fac;
    iset = TRUE;
    return (v2*fac*sd);
  }
}

static int	unitIdx ;
static int	maxUnits ;

static	void	saveOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  ((Real *)data)[unitIdx++] = unit->output ;
}

static	void	restoreOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->output = ((Real *)data)[unitIdx++] ;
}

/****************************************************************************/

void		saveOutputs(group, v)
  Group	group ;
  Real	*v;
{
  unitIdx = 0 ;
  groupForAllUnits(group, saveOutput, (void *)v) ;
}

void		restoreOutputs(group, v)
  Group	group ;
  Real	*v;
{
  unitIdx = 0 ;
  groupForAllUnits(group, restoreOutput, (void *)v) ;
}

Real		proximity(v1,v2,n)
  Real	*v1, *v2;
  int n;
{
  int i;
  Real l12 = 0.0, l1 = 0.0, l2 = 0.0;
  for (i = 0 ; i < n; i++) {
    l1  += v1[i]*v1[i];    l2  += v2[i]*v2[i];    l12 += v1[i]*v2[i];
  }
  if (l1==0.0 || l2==0.0) return 0.0;
  return (l12/sqrt(l1*l2));
}

Real		correlation(v1,v2,lv)
  Real *v1,*v2;
  int lv;
{
  int i;
  Real d1,d2,flv=((Real) lv);
  Real mean1=0.0, mean2=0.0, var1=0.0, var2=0.0, cov=0.0;
  for (i = 0; i < lv; i++) {
    mean1 += v1[i];   mean2 += v2[i];
  }
  mean1 /= flv;  mean2 /= flv;   
  for (i = 0; i < lv; i++) {
    d1 = (v1[i]-mean1);
    d2 = (v2[i]-mean2);
    var1 += d1*d1;
    var2 += d2*d2;
    cov  += d1*d2;
  }
  if (var1==0.0 || var2 == 0.0) return 0.0;
  return ((cov/flv)/(sqrt(var1/flv)*sqrt(var2/flv)));
}

Real		groupOutputProximity(group, v)
  Group	group ;
  Real	*v ;
{
  static Real	*output ;

  if (group->numUnits > maxUnits) {
    extendList((void *)&output, sizeof(Real*), maxUnits, group->numUnits) ;
    maxUnits = group->numUnits ;
  }
  
  saveOutputs(group, output) ;
  return proximity(output, v, group->numUnits) ;
}
/**********************************************************************/
static Real	alpha ;
void	initEpoch (net, onset)
  Net	net ;
  int	onset ;
{
  /* only use alpha after onset epochs */
  if (onset) {
    if (net->mz->nIterations == 0) {
      alpha = net->mz->alpha ;
      net->mz->alpha = 0 ;
      fprintf(dout,"Setting alpha to 0.0 until iteration %d\n", onset) ;
    }
    if (net->mz->nIterations == onset) {
      net->mz->alpha = alpha ;
      if (alpha != 0.0)
	fprintf(dout,"Resetting alpha from 0.0 to %f\n", alpha) ;
    }
  }

  /* this is really for checkpoint.c */
  fflush(dout) ;
}
/**********************************************************************/
int	lastTargetSlice(net)
  Net net;
{
  int	slice, groupIdx, unitIdx ;
  Group	*group ;
  Unit	*unit ;

  for (slice = net->timeSlices-1; slice >= 0; slice--) 
    for (groupIdx = 0, group = net->group ;
	 groupIdx < net->numGroups ;
	 groupIdx++, group++) 
      if ((*group)->type&OUTPUT)
	for (unitIdx = 0, unit = (*group)->unit ;
	     unitIdx < (*group)->numUnits ;
	     unitIdx++, unit++) 
	  if (!isNaN((*unit)->targetHistory[slice])) return slice+1 ;
  if (slice < 0) return net->timeSlices ;
}
/**********************************************************************/
Boolean	outputCorrect(net, tolerance, training)
  Net	net ;
  double tolerance ;
  int	training ;
{
  int	unitIdx, groupIdx ;
  Unit	*unit ;
  Group	*group ;

  if (tolerance <= 0.0) return FALSE ;

  for (groupIdx=0, group=net->group;
       groupIdx < net->numGroups; 
       groupIdx++, group++)
    if ((*group)->type&OUTPUT)
      for (unitIdx=0, unit=(*group)->unit; 
	   unitIdx < (*group)->numUnits;
	   unitIdx++, unit++) {
	Real	output = (*unit)->output ;
	Real	target = (training ? (*unit)->target : rint(output)) ;
	if (!isNaN(target) && fabs(target - output) > tolerance) 
	  return FALSE ;
      }
  return TRUE ;
}
/**********************************************************************/
int	maxOutputPosition(group)
  Group	group ;
{
  int	unitIdx, maxIdx ;
  Real	maxOutput = -1.0 ;
  Unit	*unit ;

  for (unitIdx = 0, unit = group->unit ;
       unitIdx < group->numUnits ;
       unitIdx++, unit++)
    if ((*unit)->output > maxOutput) {
      maxOutput = (*unit)->output ;
      maxIdx    = unitIdx ;
    }
  return maxIdx ;
}
/**********************************************************************/
int	cleanupWTA(group)
  Group	group ;
{
  int	unitIdx ;
  int	maxIdx = maxOutputPosition(group) ;
  Unit	*unit ;

  for (unitIdx = 0, unit = group->unit ;
       unitIdx < group->numUnits ;
       unitIdx++, unit++)
    (*unit)->output = (unitIdx == maxIdx ? 1.0 : 0.0) ;
  return maxIdx ;
}
/**********************************************************************/
void	cleanupThreshold(group, data)
  Group	group ;
  void *data ;
{
  int	unitIdx ;
  Unit	*unit ;
  Real	thresh = *(Real *)data ;

  for (unitIdx = 0, unit = group->unit ;
       unitIdx < group->numUnits ;
       unitIdx++, unit++)
    (*unit)->output = ((*unit)->output >= thresh ? 1.0 : 0.0) ;
}
/**********************************************************************/
