
/**********************************************************************
 *	   Copyright 1990,1991 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 **********************************************************************/

 /*********************************************************************
 *
 *  MFT/FEM/Boltzmann Modules written by     Evan W. Steeg
 *                                           Dept. of Computer Science
 *  August 1991                              Univ. of Toronto
 *
 **********************************************************************
 * 16dec91 (dcp)
 *	modified from mft.c
 **********************************************************************/

#include <stdio.h>
#include <math.h>
#include <values.h>
#include <nan.h>
#include <commands.h>
#include <xerion/minimize.h>
#include <xerion/utilsDCP.h>
#include "dbm.h"

struct TRACE    inRelaxation ;
struct TRACE    afterExample ;

static void	initNet     		ARGS((Net	 net)) ;
static void	deinitNet   		ARGS((Net	 net)) ;
static void	initUnit    		ARGS((Unit	 unit)) ;
static void	deinitUnit  		ARGS((Unit	 unit)) ;
static void	initLink    		ARGS((Link	 link)) ;
static void	deinitLink  		ARGS((Link	 link)) ;
static void	initExampleSet    	ARGS((ExampleSet exampleSet)) ;
static void	deinitExampleSet	ARGS((ExampleSet exampleSet)) ;

static void	calculateNetErrorDeriv  ARGS((Net net, ExampleSet exampleSet)) ;
static void	calculateNetError       ARGS((Net net, ExampleSet exampleSet)) ;
static void	netActivityUpdate	ARGS((Net   net)) ;

/***********************************************************************
 *	Name:		main 
 *	Description:	the main function, used for the xerion simulator
 *	Parameters:	
 *		int	argc	- the number of input args
 *		char	**argv  - array of argument strings from command line
 *	Return Value:	
 *		int	main	- 0
 ***********************************************************************/
int main (argc, argv)
  int	argc ;
  char	**argv ;
{
  /* Insert any private initialization routines here */
  setCreateNetHook  (initNet) ;
  setDestroyNetHook (deinitNet) ;
  setCreateUnitHook  (initUnit) ;
  setDestroyUnitHook (deinitUnit) ;
  setCreateLinkHook  (initLink) ;
  setDestroyLinkHook (deinitLink) ;
  setCreateExampleSetHook(initExampleSet) ;
  setDestroyExampleSetHook(deinitExampleSet) ;

  /* Perform initialization of the simulator */
  IStandardInit(&argc, argv);

  /* Enter loop that reads commands and handles graphics */
  ICommandLoop(stdin, stdout, NULL);

  return 0 ;
}
/***********************************************************************
 *	Name:		initNet 
 *	Description:	sets the error calculation  procedures for
 *			the net. As well, changes the activityUpdateProc,
 *			allocates memory for the extension and initializes
 *			some values in it
 *	Parameters:	
 *		Net	net - the net to act on
 *	Return Value:	NONE
 ***********************************************************************/
static void	initNet(net)
  Net	net ;
{
  net->calculateErrorDerivProc = calculateNetErrorDeriv ;
  net->calculateErrorProc      = calculateNetError ;
  net->activityUpdateProc      = netActivityUpdate ;

  net->extension = (NetExtension)calloc(1, sizeof(NetExtensionRec)) ;

  MtMax(net)    		= 5.0 ;
  MtMin(net)    		= 1.0 ;
  MtDecay(net)  		= 0.9 ;
  MinitialUnitOutput(net) 	= 0.2 ;
  MrelaxStepSize(net)		= 0.5 ;
  MrelaxTolerance(net)		= 0.01 ;
  MmaxSweepCount(net)		= 100 ;
  MtargetOffset(net)		= 0.0 ;
  MunitFunction(net)		= 0 ;		/* asymmetric [0 1] */
  MannealPositive(net)		= TRUE ;
  Mrunning(net) 		= FALSE ;
  Mcompiled(net) 		= FALSE ;
}

static void	deinitNet(net)
  Net	net ;
{
  if (net->extension != NULL)
    free(net->extension) ;
}
/***********************************************************************
 *	Name:		initUnit 
 *	Description:	allocates the memory for the unit extension record
 *	Parameters:	
 *		Unit	unit - the unit to act on
 *	Return Value:	NONE
 ***********************************************************************/
static void	initUnit(unit)
  Unit	unit ;
{
  unit->extension = (UnitExtension)calloc(1, sizeof(UnitExtensionRec)) ;
  Mlesioned(unit) = FALSE ;
}

static void	deinitUnit(unit)
  Unit	unit ;
{
  if (unit->extension != NULL)
    free(unit->extension) ;
}
/***********************************************************************
 *	Name:		initLink 
 *	Description:	allocates the memory for the link extension record
 *	Parameters:	
 *		Link	link - the link to act on
 *	Return Value:	NONE
 ***********************************************************************/
static void	initLink(link)
  Link	link ;
{
  link->extension = (LinkExtension)calloc(1, sizeof(LinkExtensionRec)) ;
  Mlesioned(link) = FALSE ;
}

static void	deinitLink(link)
  Link	link ;
{
  if (link->extension != NULL)
    free(link->extension) ;
}
/***********************************************************************
 *	Name:		initExampleSet
 *	Description:	allocates the memory for the unit extension record
 *	Parameters:	
 *		Unit	unit - the unit to act on
 *	Return Value:	NONE
 ***********************************************************************/
static void	initExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  exampleSet->extension = 
    (ExampleExtension)calloc(1, sizeof(ExampleExtensionRec)) ;
}

static void	deinitExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  if (exampleSet->extension != NULL)
    free(exampleSet->extension) ;
}

void	*getExampleSetData(exampleSet)
  ExampleSet  exampleSet ;
{
  return exampleSet->extension->data ;
}

void	putExampleSetData(exampleSet, data)
  ExampleSet  exampleSet ;
  void	*data ;
{
 exampleSet->extension->data = data ;
}
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
static	void	netCompiledSumTotalInput(net)
  Net		net ;
{
  int	idx; 
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Unit	*pre, *post ;
  Real	*wt; 

  for (idx = 0, pre=MpreUnits(net), post=MpostUnits(net), wt=Mwts(net); 
       idx < numCompiledLinks ; 
       idx++) {
    (*post)->totalInput  += (*pre)->output * (*wt) ;
    (*pre++)->totalInput += (*post++)->output * (*wt++) ;
  }
}
/**********************************************************************/
static	void	netCompiledUpdateDerivs(net)
  Net		net ;
{
  int	idx; 
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Unit	*pre, *post ;
  Real	*dv ;

  if (Mphase(net) == POSITIVE) {
    for (idx = 0, pre=MpreUnits(net), post=MpostUnits(net), dv=Mdvs(net); 
	 idx < numCompiledLinks ; 
	 idx++) 
      (*dv++) -= (*pre++)->output * (*post++)->output ;
  } else {
    for (idx = 0, pre=MpreUnits(net), post=MpostUnits(net), dv=Mdvs(net); 
	 idx < numCompiledLinks ; 
	 idx++) 
      (*dv++) += (*pre++)->output * (*post++)->output ;
  }
}
/**********************************************************************/
static	void	netCompiledUpdateCost(net)
  Net net;
{
  int	idx ;
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Real	weightCost = MweightCost(net) ;
  Real  Wo	   = MstandardWeight(net) ;
  Real	*wt, *dv ;

  if (weightCost == 0.0) return ;
  if (Wo == 0.0) {
    for (idx = 0, wt=Mwts(net), dv = Mdvs(net) ; 
	 idx < numCompiledLinks ; 
	 idx++) {
      Real wtCost = weightCost * (*wt) ;
      net->cost  += wtCost * (*wt++) ;
      (*dv++)    += 2.0 * wtCost ;
    }
  } else {
    /* WEIGHT MELTING */
    for (idx = 0, wt=Mwts(net), dv=Mdvs(net); 
	 idx < numCompiledLinks ; 
	 idx++) {
      Real ratio = (*wt) * (*wt) / (Wo * Wo) ;
      net->cost += weightCost * ratio / (1.0 + ratio) ;
      (*dv++)   += weightCost * 2.0 * (*wt++) / ((1.0 + ratio) * (1.0 + ratio)) ;
    }
  }
}
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
static void	unclamp(unit, data)
  Unit		unit ;
  void		*data ;
{
  Mclamped(unit) = FALSE ;
}
/**********************************************************************/
static Real	clampValue(net, value)
  Net	net ;
  Real	value ;
{
  Real	off	= MtargetOffset(net) ;
  Real	lo	= asym2either(net, 0.0) + off ;
  Real	hi	= asym2either(net, 1.0) - off ;

  value = asym2either(net, value) ;
  return CLIP(value, lo, hi) ;
}
/**********************************************************************/
static void	clampInput(unit, data)
  Unit		unit ;
  void		*data ;
{
  Real	sdNoise = *(Real *)data ;

  if (! isNaN(unit->output)) {
    unit->output = clampValue(unit->net, unit->output) ;
    if (sdNoise != 0.0) unit->output += gaussianNoise(sdNoise) ;
    Mclamped(unit) = TRUE ;
  }  
}
/**********************************************************************/
  static void	clampTarget(unit, data)
  Unit		unit ;
  void		*data ;
{
  if (! isNaN(unit->target)) {
    unit->output = clampValue(unit->net, unit->target) ;
    Mclamped(unit) = TRUE ;
  }  
}
/**********************************************************************/
static void	initOutput(unit, data)
  Unit		unit ;
  void		*data ;
{
  if (!Mclamped(unit)) 
    unit->output = *(Real *)data ;
}
/**********************************************************************/
static void	zeroTotalInput(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->totalInput = 0.0 ;
}
/**********************************************************************/
static void	computeOutput(unit, data)
  Unit		unit ;
  void		*data ;
{
  Net	net 	= unit->net ;
  Real	relax	= *(Real *)data ;
  Real	input, output, delta ;

  if (Mclamped(unit)) return ;
  input = unit->totalInput/Mtemp(net) + gaussianNoise(MunitNoise(net)) ;
  output = sigmoid(input) ;
  output = asym2either(net, output) ;
  delta = relax * (output - unit->output) ;
  MmaxDelta(net) = MAX(MmaxDelta(net), fabs(delta)) ;
  unit->output  += delta ;
}
/***********************************************************************/
static Real	unitError(unit)
  const Unit	unit ;
{
  Net	net	= unit->net ;
  Real	off	= MtargetOffset(net) ;
  Real	target	= unit->target ;
  Real	output 	= unit->output ;
  Real	error ;

  if (isNaN(target)) return 0.0 ;

  target = CLIP(target, off, 1.0-off) ;
  output = either2asym(net, output) ;
  output = CLIP(output, MIN_OUTPUT, MAX_OUTPUT) ;

  if (MzeroErrorRadius(net) <= 0.0) {
    error = INFO(output, target) ;
  } else {
    Real	upper = target + MzeroErrorRadius(net) ;
    Real	lower = target - MzeroErrorRadius(net) ;
    upper = MIN(1.0, upper) ;
    lower = MAX(0.0, lower) ;
    error = (output > upper ? INFO(output, upper) : 
	     (output < lower ? INFO(output, lower) : 
	      0.0));
  }
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  return error ;
}
/**********************************************************************/
static void	incrementError(unit, data)
  Unit		unit ;
  void		*data ;
{
  *(Real *)data += unitError(unit) ;
}
/**********************************************************************/
Real		groupError(group)
  Group group ;
{
  Real	error = 0 ;
  
  groupForAllUnits(group, incrementError, &error) ;
  return error ;
}
/***********************************************************************/
static void	netAnneal(net)
  Net		net ;
{
  int	sweepCount 	= 0 ;
  int	maxCount	= MmaxSweepCount(net) ;
  Boolean	anneal	= (Mphase(net) == NEGATIVE || MannealPositive(net)) ;
  Real		temp	= (anneal ? MtMax(net) - MtMin(net) : 0.0) ;
  Real		relax	= (anneal ? MrelaxStepSize(net) : 1.0) ;

  do {
    Mtemp(net) = MtMin(net) + temp ;
    MmaxDelta(net) = 0.0;
    
    netForAllUnits     (net, ALL,  zeroTotalInput, NULL) ;
    netCompiledSumTotalInput(net) ;
    netForAllOtherUnits(net, BIAS, computeOutput,  (void *)&relax) ;

    if (Mrunning(net) != TRUE) {
      int	idx, delayCount = MdelayCount(net) ;
      IDoTrace(&inRelaxation) ;
      /* For viewing relaxation */
      for (idx = 0 ; idx < delayCount*1000 ; idx++) ;
    }
    sweepCount++ ;
    temp *= MtDecay(net) ;

    if (Mdebug(net)) {
      fprintf(dout, "%d %f %f\n", sweepCount, Mtemp(net), MmaxDelta(net)) ;
      if (!anneal)
	fprintf(dout, "not annealing POSITIVE\n") ;
      else if (MmaxDelta(net) <= MrelaxTolerance(net))
	fprintf(dout, "maxDelta %f <= tol %f\n",
		MmaxDelta(net), MrelaxTolerance(net)) ;
      else if (sweepCount == maxCount)
	fprintf(dout, "max sweep count %d\n", sweepCount) ;
    }

  } while (anneal				  &&
	   (!maxCount || (sweepCount < maxCount)) &&
	   (MmaxDelta(net) > MrelaxTolerance(net))) ;
  MaveSweepCount(net) += sweepCount;
}
/***********************************************************************/
static void	netActivityUpdate(net)
  Net		net ;
{
  Real	init	= asym2either(net, MinitialUnitOutput(net)) ;
  Real	sdNoise	= MinputNoise(net) ;

  if (!Mrunning(net)) {
    Mphase(net) 		= NEGATIVE ;
    net->error 			= 0.0 ;
    MexampleFrequency(net)	= 1.0 ;
    MaveSweepCount(net)		= 0 ;
    getCompiledWeights(net) ;
  }
  netForAllUnits     (net, INPUT|OUTPUT, unclamp, 	 NULL) ;
  netForAllUnits     (net, INPUT, 	 clampInput,  	 (void *)&sdNoise) ;
  if (Mphase(net) == POSITIVE)
    netForAllUnits   (net, OUTPUT,     	 clampTarget, 	 NULL) ;
  netForAllOtherUnits(net, BIAS,    	 initOutput,  	 (void *)&init) ;

  netAnneal(net) ;

  if (Mphase(net) == NEGATIVE) 
    netForAllUnits   (net, OUTPUT,     	 incrementError, (void *)&net->error) ;

  netCompiledUpdateDerivs(net) ;

  if (!Mrunning(net)) {
    putCompiledDerivs(net) ;	/* just to be able to look at them */
  }
}
/***********************************************************************/
static	Boolean	getExample(net, exampleSet, next) 
  Net		net ;
  ExampleSet	exampleSet ;
  int		next ;
{
  if (MbatchSize(net) < exampleSet->numExamples) 
    /* select random example according to it's probability */
    do {
      MexampleNum(net) = randomInteger(exampleSet->numExamples);
      MexampleFrequency(net) = exampleSet->example[MexampleNum(net)].frequency ;
    } while (MsampleExamples(net) &&
	     !WITH_PROBABILITY(MexampleFrequency(net))) ;
  else {
    MexampleNum(net) = next ;
    MexampleFrequency(net) = exampleSet->example[next].frequency ;
    if (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net)))
      /* skip this example */
      return FALSE ;
  }
  MgetNumber(exampleSet, MexampleNum(net)) ;
  return TRUE ;
}
/***********************************************************************/
static void	calculateNetErrorDeriv(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  int	numExamples ;

  Mrunning(net) 	= TRUE ;
  net->error 		= 0.0 ;
  MaveSweepCount(net) 	= 0 ;
  getCompiledWeights(net) ;
  initEpoch(net, MalphaOnset(net)) ;

  for (numExamples = 0 ; numExamples < MbatchSize(net) ; ++numExamples) {
    if (!getExample(net, exampleSet, numExamples)) continue ;

    Mphase(net) = POSITIVE ;
    netActivityUpdate(net) ;

    Mphase(net) = NEGATIVE ;
    netActivityUpdate(net) ;

    IDoTrace(&afterExample) ;
  }
  if (numExamples <= 0)
    IErrorAbort("calculateNetErrorDeriv: no examples processed") ;
  else 
    MaveSweepCount(net) /= numExamples ;

  /* Update the cost after all examples done.  */
  net->cost = 0.0 ;
  netCompiledUpdateCost(net) ;

  putCompiledDerivs(net) ;
  Mrunning(net) = FALSE ;
}
/***********************************************************************/
static void	calculateNetError(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  int	numExamples ;

  Mrunning(net) 	= TRUE ;
  MaveSweepCount(net) 	= 0 ;
  net->error 		= 0.0 ;
  getCompiledWeights(net) ;

  for (numExamples = 0 ; numExamples < MbatchSize(net) ; ++numExamples) {
    if (!getExample(net, exampleSet, numExamples)) continue ;

    Mphase(net) = NEGATIVE ;
    netActivityUpdate(net) ;

    IDoTrace(&afterExample) ;
  }
  if (numExamples <= 0)
    IErrorAbort("calculateNetError: no examples processed") ;
  else 
    MaveSweepCount(net) /= numExamples ;

  /* Update the cost after all examples done.  */
  net->cost = 0.0 ;
  netCompiledUpdateCost(net) ;

  Mrunning(net) = FALSE ;
}
