/**********************************************************************
 * $Id: bp.c,v 1.9 91/09/26 11:05:14 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	  Copyright 1990,1991 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <math.h>
#include <values.h>
#include <nan.h>
#include <commands.h>
#include <xerion/minimize.h>
#include <xerion/utilsDCP.h>
#include "elman.h"

struct TRACE    inRelaxation ;
struct TRACE	forwardEachTime ;
struct TRACE    afterExample ;

static Proc	updateNetActivity      ARGS((Net	net)) ;
static void	calculateNetErrorDeriv ARGS((Net net, ExampleSet exampleSet)) ;
static void	calculateNetError      ARGS((Net net, ExampleSet exampleSet)) ;

static void	initNet   		ARGS((Net	 net)) ;
static void	deinitNet 		ARGS((Net	 net)) ;
static void	initUnit    		ARGS((Unit	 unit)) ;
static void	deinitUnit  		ARGS((Unit	 unit)) ;
static void	initLink    		ARGS((Link	 link)) ;
static void	deinitLink  		ARGS((Link	 link)) ;
static void	initExampleSet    	ARGS((ExampleSet exampleSet)) ;
static void	deinitExampleSet	ARGS((ExampleSet exampleSet)) ;

/**********************************************************************
 *	Name:		main 
 *	Description:	the main function, used for the xerion simulator
 *	Parameters:	
 *		int	argc	- the number of input args
 *		char	**argv  - array of argument strings from command 
 *				  line
 *	Return Value:	
 *		int	main	- 0
 ***********************************************************************/
int main (argc, argv)
  int	argc ;
  char	**argv ;
{
  /* Insert any private initialization routines here */
  setCreateNetHook(initNet) ;
  setDestroyNetHook(deinitNet) ;
  setCreateUnitHook(initUnit) ;
  setDestroyUnitHook(deinitUnit) ;
  setCreateLinkHook(initLink) ;
  setDestroyLinkHook(deinitLink) ;
  setCreateExampleSetHook(initExampleSet) ;
  setDestroyExampleSetHook(deinitExampleSet) ;

  /* Perform initialization of the simulator */
  IStandardInit(&argc, argv);

  /* Enter loop that reads commands and handles graphics */
  ICommandLoop(stdin, stdout, NULL);

  return 0 ;
}
/***********************************************************************/
static void	initNet (net)
  Net	net ;
{
  net->calculateErrorDerivProc = calculateNetErrorDeriv ;
  net->calculateErrorProc      = calculateNetError ;
  net->activityUpdateProc      = updateNetActivity ;

  net->extension = (NetExtension)calloc(1, sizeof(NetExtensionRec)) ;

  Mcompiled(net)           	= FALSE ;
  MsampleExamples(net)		= FALSE ;
  MzeroErrorRadius(net)   	= 0.1 ;
  McorrectTolerance(net) 	= 0.2 ;
  MinitialUnitOutput(net)	= 0.2 ;
  MinputNoise(net)		= 0.0 ;
  MrelaxStepSize(net)		= 1.0 ;
  MmaxRelaxSteps(net)		= 1 ;
  MnumSlices(net)	   	= 0 ;
}
/**********************************************************************/
static void	deinitNet (net)
  Net	net ;
{
  if (net->extension != NULL)
    free(net->extension) ;
}
/***********************************************************************/
static void	initUnit(unit)
  Unit	unit ;
{
  unit->extension = (UnitExtension)calloc(1, sizeof(UnitExtensionRec)) ;
  Mlesioned(unit)        = FALSE ;
  MpreOutputs(unit)	 = NULL ;
  MpreOutputDerivs(unit) = NULL ;
  Mweights(unit)    	 = NULL ;
  Mderivs(unit)     	 = NULL ;
}

static void	deinitUnit(unit)
  Unit	unit ;
{
  if (unit->extension != NULL)
    free(unit->extension) ;
}
/***********************************************************************/
static void	initLink(link)
  Link	link ;
{
  link->extension = (LinkExtension)calloc(1, sizeof(LinkExtensionRec)) ;
  Mlesioned(link) = FALSE ;
}

static void	deinitLink(link)
  Link	link ;
{
  if (link->extension != NULL)
    free(link->extension) ;
}
/***********************************************************************/
static void	initExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  exampleSet->extension = 
    (ExampleExtension)calloc(1, sizeof(ExampleExtensionRec)) ;
}

static void	deinitExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  if (exampleSet->extension != NULL)
    free(exampleSet->extension) ;
}

void	*getExampleSetData(exampleSet)
  ExampleSet  exampleSet ;
{
  return exampleSet->extension->data ;
}

void	putExampleSetData(exampleSet, data)
  ExampleSet  exampleSet ;
  void	*data ;
{
 exampleSet->extension->data = data ;
}
/**********************************************************************/
/**********************************************************************/
static Real	unitError(unit)
  const Unit	unit ;
{
  Net	net = unit->net ;
  Real	error, output, target = unit->target ;

  if (isNaN(target)) return 0.0 ;
  output = CLIP(unit->output, MIN_OUTPUT, MAX_OUTPUT) ;

  if (MzeroErrorRadius(net) <= 0.0) {
    error = INFO(output, target) ;
  } else {
    Real	upper = target + MzeroErrorRadius(net) ;
    Real	lower = target - MzeroErrorRadius(net) ;
    upper = MIN(1.0, upper) ;
    lower = MAX(0.0, lower) ;
    error = (output > upper ? INFO(output, upper) : 
	     (output < lower ? INFO(output, lower) : 
	      0.0));
  }
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  return error ;
}
/**********************************************************************/
static Real	unitErrorDeriv(unit)
  const Unit	unit ;
{
  Real	error, output, target = unit->target ;
  Net	net = unit->net ;

  if (isNaN(target)) return 0.0 ;
  output = CLIP(unit->output, MIN_OUTPUT, MAX_OUTPUT) ;

  if (MzeroErrorRadius(net) <= 0.0) {
    error = INFO_DERIV(output, target) ;
  } else {
    Real	upper = target + MzeroErrorRadius(net) ;
    Real	lower = target - MzeroErrorRadius(net) ;
    upper = MIN(1.0, upper) ;
    lower = MAX(0.0, lower) ;
    error = (output > upper ? INFO_DERIV(output, upper) : 
	     (output < lower ? INFO_DERIV(output, lower) : 
	      0.0));
  }
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  return error ;
}
/**********************************************************************/
static void	incrementError(unit, data)
  Unit		unit ;
  void		*data ;
{
  *(Real *)data += unitError(unit) ;
}

Real		groupError(group)
  Group group ;
{
  Real	error = 0 ;
  groupForAllUnits(group, incrementError, &error) ;
  return error ;
}
/**********************************************************************/
static void	injectError(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->outputDeriv = (unit->group->type&OUTPUT ? unitErrorDeriv(unit) : 0.0) ;
}
/**********************************************************************/
static void	invertError(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->inputDeriv = sigmoidDeriv(unit->output) * unit->outputDeriv ;
}
/**********************************************************************/
static void	updateCost(unit, data)
  Unit	unit ;
  void	*data ;
{
  Net	net = (Net)data ;
  int	numIncoming = unit->numIncoming ;
  Real	*wt, *dv ;
  Real	weightCost = MweightCost(net) ;
  Real	Wo	   = MstandardWeight(net) ;
  int	idx ;

  if (weightCost == 0.0) return ;
  if (Wo == 0.0) {
    /* WEIGHT DECAY */
    for (idx=0, wt=Mweights(unit), dv=Mderivs(unit) ;
	 idx < numIncoming ;
	 idx++) {
      Real  wtCost = (*wt) * weightCost ;
      net->cost   += wtCost * (*wt++) ;
      (*dv++)     += 2.0 * wtCost ;
    }
  } else {
    /* WEIGHT MELTING */
    for (idx=0, wt=Mweights(unit), dv=Mderivs(unit) ;
	 idx < numIncoming ;
	 idx++) {
      Real ratio = (*wt) * (*wt) / (Wo * Wo) ;
      net->cost += weightCost * ratio / (1.0+ratio) ;
      (*dv++)   += weightCost * 2.0 * (*wt++) / ((1.0+ratio)*(1.0+ratio)) ;
    }
  }
}
/**********************************************************************/
/**********************************************************************/
static void	setInput(unit, time)
  Unit		unit ;
  void		*time ;
{
  unit->output = unit->outputHistory[(int)time] =
    fabs(unit->clampingHistory[(int)time] - 
	 fabs(gaussianNoise(MinputNoise(unit->net)))) ;
}
/**********************************************************************/
static void	setTarget(unit, time)
  Unit		unit ;
  void		*time ;
{
  unit->target = unit->targetHistory[(int)time] ;
}
/**********************************************************************/
static void	setOutput(unit, output)
  Unit		unit ;
  void		*output ;
{
  unit->output = *(Real *)output ;
}
/**********************************************************************/
static void	forceTarget(unit, data)
  Unit		unit ;
  void		*data ;
{
  if (!isNaN(unit->target))
    unit->output = unit->target ;
}
/**********************************************************************/
static void	saveOutput(unit, time)
  Unit		unit ;
  void		*time ;
{
  MoldOutput(unit) = unit->outputHistory[(int)time] = unit->output ;
}
/**********************************************************************/
static void	restoreOldOutput(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->output = MoldOutput(unit) ;
}
/**********************************************************************/
static void	restoreOutputFromTime(unit, time)
  Unit		unit ;
  void		*time ;
{
  unit->output = unit->outputHistory[(int)time] ;
}
/***********************************************************************/
static	void	netPresentExample(net)
  Net	net ;
{
  netForAllUnits(net, INPUT,  setInput,  (void *)Mtime(net)) ;
  netForAllUnits(net, OUTPUT, setTarget, (void *)Mtime(net)) ;
}
/**********************************************************************/
/**********************************************************************/
static	void	forward(unit, time)
  Unit	unit ;
  void	*time ;
{
  int	numIncoming 	= unit->numIncoming ;
  Real	relaxStep	= MrelaxStepSize(unit->net) ;
  Real	totalInput  	= 0.0 ;
  Real	*wt, **preOutput ;
  int	idx ;

  for (idx=0, wt=Mweights(unit), preOutput=MpreOutputs(unit) ;
       idx < numIncoming ; 
       idx++) 
    totalInput += (*wt++) * *(*preOutput++) ;
  unit->output += relaxStep * (sigmoid(totalInput) - unit->output) ;
  unit->outputHistory[(int)time] = unit->output ;
}
/***********************************************************************/
static void	netForward (net)
  Net		net ;
{
  /* restore states as "input" units and then compute new states */
  netForAllOtherUnits(net, INPUT|BIAS, restoreOldOutput, NULL) ;
  netForAllOtherUnits(net, INPUT|BIAS, forward,          (void *)Mtime(net)) ;

  if (!Mrunning(net)) {
    int	idx, delayCount = MdelayCount(net) ;
    IDoTrace(&inRelaxation) ;
    /* For viewing relaxation */
    for (idx = 0 ; idx < delayCount*1000 ; idx++) ;
  }
}
/**********************************************************************/
/**********************************************************************/
static	void	backward(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	numIncoming 	= unit->numIncoming ;
  Real	inputDeriv	= unit->inputDeriv ;
  Real 	*wt, *dv, **preOut, **preDv ;
  int	idx ;

  for (idx=0, wt=Mweights(unit), dv=Mderivs(unit), preOut=MpreOutputs(unit),
       preDv=MpreOutputDerivs(unit) ;
       idx < numIncoming ; 
       idx++) {
    *(*preDv++) += inputDeriv *  (*wt++) ;
     (*dv++)    += inputDeriv * *(*preOut++) ;
  }
}
/**********************************************************************/
static	void	backwardGroup(group, time)
  Group	group ;
  void	*time;
{
  groupForAllUnits(group, invertError,      NULL) ;
  groupForAllUnits(group, restoreOldOutput, NULL) ;
  groupForAllUnits(group, backward,         NULL) ;
}
/***********************************************************************/
static void	netBackward (net)
  Net		net ;
{
  netForAllOtherUnits     (net, INPUT|BIAS, injectError,         NULL) ;
  netForAllOtherGroupsBack(net, INPUT|BIAS, backwardGroup,       (void *)Mtime(net)) ;
  netForAllOtherUnits     (net, INPUT|BIAS, restoreOutputFromTime, (void *)Mtime(net)) ;
}
/**********************************************************************/
/***********************************************************************/
static	void	clearStats(net)
  Net	net ;
{
  if (net->mz->repetitionCount==0 || net->mz->nIterations==0 ||
      (net->mz->nIterations-1)%net->mz->repetitionCount==0) {
    net->error 		   	= 0 ;
    net->cost 		   	= 0 ;
    MaveRelaxSteps(net)	   	= 0 ;
    MaveFixations(net) 	   	= 0 ;
    MavePhonemesCorrect(net)	= 0 ;
    MaveWordsCorrect(net)    	= 0 ;
  }
}
/***********************************************************************/
static	void	normStats(net)
  Net	net ;
{
  Real	scale = 1.0 ;

  if (net->mz->repetitionCount==0 || 
      net->mz->nIterations%net->mz->repetitionCount==0) {
    if (net->mz->nIterations && net->mz->repetitionCount)
      scale = (Real)net->mz->repetitionCount ;
    net->error		     /= scale ;
    net->cost		     /= scale ;
    scale *= net->batchSize ;
    MaveRelaxSteps(net)	     /= scale ;
    MaveFixations(net)	     /= scale ;
    MavePhonemesCorrect(net) /= scale ;
    MaveWordsCorrect(net)    /= scale ;
  }
}
/***********************************************************************/
/***********************************************************************/
static void	netIterate(net, training)
  Net		net ;
  Boolean	training ;
{
  int	maxTime = MnumSlices(net) = lastTargetSlice(net) ;
  int	relaxIdx ;

  netForAllOtherUnits(net, INPUT|BIAS, setOutput, (void *)&MinitialUnitOutput(net)) ;

  for (Mtime(net) = 0 ; Mtime(net) < maxTime ; Mtime(net)++) {
    netPresentExample(net) ;
    netForAllOtherUnits(net, INPUT|BIAS, saveOutput, (void *)Mtime(net)) ;

    for (relaxIdx = 0 ; relaxIdx < MmaxRelaxSteps(net) ; relaxIdx++)
      netForward(net) ;

    netForAllUnits(net, OUTPUT,	incrementError, (void *)&(net->error)) ;
    if (training) netBackward(net) ;
    if (MteacherForcing(net)) netForAllUnits(net, OUTPUT, forceTarget, NULL) ;
    IDoTrace(&forwardEachTime) ;
  }
}

/***********************************************************************/
static void	netIterateRefixate(net, training)
  Net		net ;
  Boolean	training ;
{
  int	relaxIdx, numPhonemesCorrect=0 ;
  int	maxTime = MnumSlices(net) = lastTargetSlice(net) ;
  int	fixation=0, attention=0, saccade=0, phonemesCorrect=FALSE ; 
  Real	thresh = 0.5 ;
  Group	Letters, Attention, Phonemes ;

  if ((Letters=groupFromName(net, "Letters")) == NULL) 
    IErrorAbort("no \"Letters\" group") ;
  if ((Attention=groupFromName(net, "Attention")) == NULL) 
    IErrorAbort("no \"Attention\" group") ;
  if ((Phonemes=groupFromName(net, "Phonemes")) == NULL) 
    IErrorAbort("no \"Phonemes\" group") ;
  
  netForAllOtherUnits(net, INPUT|BIAS, setOutput, (void *)&MinitialUnitOutput(net)) ;
  MaveFixations(net)++ ;

  for (Mtime(net) = 0 ; Mtime(net) < maxTime ; Mtime(net)++) {
    Fixate(Letters, Attention, saccade, attention, Mtime(net)) ;
    netPresentExample(net) ;
    netForAllOtherUnits(net, INPUT|BIAS, saveOutput, (void *)Mtime(net)) ;

    for (relaxIdx = 0 ; relaxIdx < MmaxRelaxSteps(net) ; relaxIdx++) {
      netForward(net) ;
      MaveRelaxSteps(net)++ ;
      phonemesCorrect = outputCorrect(net, McorrectTolerance(net), training) ;
      if (phonemesCorrect) break ;
    }
    netForAllUnits(net, OUTPUT,	incrementError, (void *)&(net->error)) ;
    if (training) netBackward(net) ;
    if (!Mrunning(net)) IDoTrace(&forwardEachTime) ;

    if (Mdebug(net)) fprintf(dout, "%d/%d\tfixate %d(%d)\t", Mtime(net), relaxIdx, fixation, attention) ;

    /* CORRECT - TRY NEXT PHONEME WITHOUT REFIXATING */
    if (phonemesCorrect) {
      if (Mdebug(net)) fprintf(dout, "correct %s - no saccade\n", (attention ? "in periphery" : "at fixation ")) ;
      saccade   = fixation ;	/* no change */
/*
      if (MteacherForcing(net)) {
	if (training) netForAllUnits(net, OUTPUT, forceTarget, NULL) ;
	else netForAllGroups(net, OUTPUT, cleanupThreshold, (void *)&thresh) ;
      }
*/
      attention = maxOutputPosition(Attention) ;
      numPhonemesCorrect++ ;

    /* FAILED ON FIXATED GRAPHEME - REFIXATE NEXT GRAPHEME */
    } else if (attention == 0) { 
      if (MteacherForcing(net)) {
	if (training) netForAllUnits(net, OUTPUT, forceTarget, NULL) ;
/*
	else netForAllGroups(net, OUTPUT, cleanupThreshold, (void *)&thresh) ;
*/
      }
      attention = maxOutputPosition(Attention) ;
      netForAllOtherUnits(net, INPUT|OUTPUT|BIAS, setOutput, (void *)&MinitialUnitOutput(net));
      if (Mdebug(net)) fprintf(dout, "fail    at fixation  - saccade %d\n", attention) ;
      fixation += attention ;
      saccade   = fixation ;
      attention = 0 ;
      MaveFixations(net)++ ;

    /* FAILED ON NONFIXATED GRAPHEME - REFIXATE IT AND TRY AGAIN */
    } else {
      Mtime(net)-- ;
      netForAllOtherUnits(net, INPUT|BIAS, restoreOutputFromTime, (void *)Mtime(net)) ;
      if (Mdebug(net)) fprintf(dout, "fail    in periphery - saccade %d and repeat\n", attention) ;
      fixation += attention ;
      saccade   = attention ;		/* applied to OLD states/targets */
      attention = 0 ;
      MaveFixations(net)++ ;
    } 
  } /* time */
  MavePhonemesCorrect(net) += numPhonemesCorrect ;
  if (numPhonemesCorrect == maxTime) MaveWordsCorrect(net)++ ;
}
/***********************************************************************/
static	Boolean	getExample(net, exampleSet, next) 
  Net		net ;
  ExampleSet	exampleSet ;
  int		next ;
{
  if (MbatchSize(net) < exampleSet->numExamples) 
    /* select random example according to it's probability */
    do {
      MexampleNum(net) = randomInteger(exampleSet->numExamples);
      MexampleFrequency(net) = exampleSet->example[MexampleNum(net)].frequency ;
    } while (MsampleExamples(net) &&
	     !WITH_PROBABILITY(MexampleFrequency(net))) ;
  else {
    MexampleNum(net) = next ;
    MexampleFrequency(net) = exampleSet->example[next].frequency ;
    if (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net)))
      /* skip this example */
      return FALSE ;
  }
  MgetNumber(exampleSet, MexampleNum(net)) ;
  return TRUE ;
}
/**********************************************************************/
static Proc	updateNetActivity (net)
  Net	net ;
{
  Mrunning(net) = FALSE ;
  if (!Mcompiled(net)) netCompile(net) ;
  net->error = 0.0 ;
  MexampleFrequency(net)   = 1.0 ;
  MaveRelaxSteps(net)	   = 0 ;
  MaveFixations(net)       = 0 ;
  MavePhonemesCorrect(net) = 0 ;
  MaveWordsCorrect(net)    = 0 ;
  netForAllOtherUnits(net, INPUT|BIAS, getCompiledWeights, NULL) ;

  if (Mrefixate(net)) netIterateRefixate(net, FALSE) ;
  else		      netIterate(net, FALSE) ;

  /* restore time to maxTime for display purposes */
  Mtime(net) = MnumSlices(net) - 1 ;
  IDoTrace(&afterExample) ;
}
/***********************************************************************/
static void	calculateError(net, exampleSet, calculateDerivs)
  Net		net ;
  ExampleSet	exampleSet ;
  Boolean	calculateDerivs ;
{
  int 	numExamples ;

  Mrunning(net) = TRUE ;
  if (!Mcompiled(net)) netCompile(net) ;
  if (calculateDerivs) initEpoch(net, MalphaOnset(net)) ;
  clearStats(net) ;
  netForAllOtherUnits(net, INPUT|BIAS, getCompiledWeights, NULL) ;

  for (numExamples = 0 ; numExamples < MbatchSize(net) ; ++numExamples) {
    if (!getExample(net, exampleSet, numExamples)) continue ;
    if (Mrefixate(net)) netIterateRefixate(net, calculateDerivs) ;
    else		netIterate(net, calculateDerivs) ;
    IDoTrace(&afterExample) ;
  }

  if (numExamples <= 0)
    IErrorAbort("calculateError: no examples processed") ;
  netForAllUnits(net, ALL, updateCost, net) ;
  if (calculateDerivs) 
    netForAllOtherUnits(net, INPUT|BIAS, putCompiledDerivs, NULL) ;
  normStats(net) ;
  Mrunning(net) = FALSE ;
}
/***********************************************************************/
static void	calculateNetErrorDeriv(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  calculateError(net, exampleSet, TRUE) ;
}
/*********************************************************************/
static void	calculateNetError(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  calculateError(net, exampleSet, FALSE) ;
}
/*********************************************************************/
