
/**********************************************************************
 * $Id: bp.c,v 1.9 91/09/26 11:05:14 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	  Copyright 1990,1991 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <math.h>
#include <values.h>
#include <nan.h>
#include <commands.h>
#include <xerion/minimize.h>
#include <xerion/utilsDCP.h>
#include "ffbp.h"

struct TRACE    afterExample ;

static void	forward      ARGS((Unit	unit, void *data)) ;
static void	backward     ARGS((Unit	unit, void *data)) ;

static void	netForward  ARGS((Net	 net)) ;
static void	netBackward ARGS((Net	 net)) ;

static Proc	updateNetActivity      ARGS((Net	net)) ;
static Proc	updateNetGradient      ARGS((Net	net)) ;

static void	calculateNetErrorDeriv ARGS((Net net, ExampleSet exampleSet)) ;
static void	calculateNetError      ARGS((Net net, ExampleSet exampleSet)) ;

static void	initNet   		ARGS((Net	 net)) ;
static void	deinitNet 		ARGS((Net	 net)) ;
static void	initUnit    		ARGS((Unit	 unit)) ;
static void	deinitUnit  		ARGS((Unit	 unit)) ;
static void	initLink    		ARGS((Link	 link)) ;
static void	deinitLink  		ARGS((Link	 link)) ;
static void	initExampleSet    	ARGS((ExampleSet exampleSet)) ;
static void	deinitExampleSet	ARGS((ExampleSet exampleSet)) ;

/**********************************************************************
 *	Name:		main 
 *	Description:	the main function, used for the xerion simulator
 *	Parameters:	
 *		int	argc	- the number of input args
 *		char	**argv  - array of argument strings from command 
 *				  line
 *	Return Value:	
 *		int	main	- 0
 ***********************************************************************/
int main (argc, argv)
  int	argc ;
  char	**argv ;
{
  /* Insert any private initialization routines here */
  setCreateNetHook(initNet) ;
  setDestroyNetHook(deinitNet) ;
  setCreateUnitHook(initUnit) ;
  setDestroyUnitHook(deinitUnit) ;
  setCreateLinkHook(initLink) ;
  setDestroyLinkHook(deinitLink) ;
  setCreateExampleSetHook(initExampleSet) ;
  setDestroyExampleSetHook(deinitExampleSet) ;

  /* Perform initialization of the simulator */
  IStandardInit(&argc, argv);

  /* Enter loop that reads commands and handles graphics */
  ICommandLoop(stdin, stdout, NULL);

  return 0 ;
}
/***********************************************************************/
static void	initNet (net)
  Net	net ;
{
  net->calculateErrorDerivProc = calculateNetErrorDeriv ;
  net->calculateErrorProc      = calculateNetError ;
  net->activityUpdateProc      = updateNetActivity ;
  net->gradientUpdateProc      = updateNetGradient ;

  net->extension = (NetExtension)calloc(1, sizeof(NetExtensionRec)) ;

  MzeroErrorRadius(net)	= 0.2 ;
  MsampleExamples(net)	= FALSE ;
  MweightCostIncrement(net)	= 0.0 ;
  MweightCostDecrement(net)	= 0.9 ;
  MaveErrorFactor(net)		= 0.95 ;
  MdesiredError(net)		= 0.0 ;
}
static void	deinitNet (net)
  Net	net ;
{
  if (net->extension != NULL)
    free(net->extension) ;
}
/***********************************************************************/
static void	initUnit(unit)
  Unit	unit ;
{
  unit->extension = (UnitExtension)calloc(1, sizeof(UnitExtensionRec)) ;
  Mlesioned(unit) 	 = FALSE ;
  MpreOutputs(unit)	 = NULL ;
  MpreOutputDerivs(unit) = NULL ;
  Mweights(unit)    	 = NULL ;
  Mderivs(unit)     	 = NULL ;
}

static void	deinitUnit(unit)
  Unit	unit ;
{
  if (unit->extension != NULL)
    free(unit->extension) ;
}
/***********************************************************************/
static void	initLink(link)
  Link	link ;
{
  link->extension = (LinkExtension)calloc(1, sizeof(LinkExtensionRec)) ;
  Mlesioned(link) = FALSE ;
}

static void	deinitLink(link)
  Link	link ;
{
  if (link->extension != NULL)
    free(link->extension) ;
}
/***********************************************************************
 *	Name:		initExampleSet
 *	Description:	allocates the memory for the unit extension record
 *	Parameters:	
 *		Unit	unit - the unit to act on
 *	Return Value:	NONE
 ***********************************************************************/
static void	initExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  exampleSet->extension = 
    (ExampleExtension)calloc(1, sizeof(ExampleExtensionRec)) ;
}

static void	deinitExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  if (exampleSet->extension != NULL)
    free(exampleSet->extension) ;
}

void	*getExampleSetData(exampleSet)
  ExampleSet  exampleSet ;
{
  return exampleSet->extension->data ;
}

void	putExampleSetData(exampleSet, data)
  ExampleSet  exampleSet ;
  void	*data ;
{
 exampleSet->extension->data = data ;
}
/**********************************************************************/
/***********************************************************************
 *	Name:		unitError
 *	Description:	calculates the error value of a unit assuming
 *			a zeroErrorRadius
 *	Parameters:	
 *		const Unit	unit - the unit to calclate the error of
 *	Return Value:	
 *		Real		unitError - the error of the unit
 ***********************************************************************/
static Real	unitError(unit)
  const Unit	unit ;
{
  Net	net = unit->net ;
  Real	error, output = unit->output, target = unit->target ;

  if (isNaN(target) || fabs(output-target) <= MzeroErrorRadius(net))
    return 0.0 ;

  switch (MerrorFunction(net)) {
  case CROSS_ENTROPY:
    output = CLIP(output, MIN_OUTPUT, MAX_OUTPUT) ;
    error = INFO(output, target) ;
    break ;
  case SQUARED_ERROR:
    error = (output - target) * (output - target) ;
    break ;
  default:
    IErrorAbort("Unknown error function: %d", MerrorFunction(net)) ;
  }      
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  return error ;
}
/**********************************************************************/
static Real	unitErrorDeriv(unit)
  const Unit	unit ;
{
  Net	net = unit->net ;
  Real	error, output = unit->output, target = unit->target ;

  if (isNaN(target) || fabs(output-target) <= MzeroErrorRadius(net))
    return 0.0 ;

  switch (MerrorFunction(net)) {
  case CROSS_ENTROPY:
    output = CLIP(output, MIN_OUTPUT, MAX_OUTPUT) ;
    error = INFO_DERIV(output, target) ; 
    break ;
  case SQUARED_ERROR:
    error = 2.0 * (output - target) ;
    break ;
  default:
    IErrorAbort("Unknown error function: %d", MerrorFunction(net)) ;
  }      
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  return error ;
}
/**********************************************************************/
static void	incrementError(unit, data)
  Unit		unit ;
  void		*data ;
{
  *(Real *)data += unitError(unit) ;
}

Real		groupError(group)
  Group group ;
{
  Real	error = 0 ;
  groupForAllUnits(group, incrementError, &error) ;
  return error ;
}
/**********************************************************************/
static void	injectError(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->outputDeriv = (unit->group->type&OUTPUT ? unitErrorDeriv(unit) : 0.0) ;
}
/**********************************************************************/
static void	incrementCost(unit, data)
  Unit		unit ;
  void		*data ;
{
  *(Real *)data += MunitCost(unit->net) * unit->output * (1.0 - unit->output) ;
}

static void	injectCost(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->outputDeriv += MunitCost(unit->net) * (1.0 - 2.0 * unit->output) ;
}
/**********************************************************************/
static void	updateCost(unit, data)
  Unit	unit ;
  void	*data ;
{
  Net	net = (Net)data ;
  int	numIncoming = unit->numIncoming ;
  Real	*wt, *dv ;
  Real	weightCost = MweightCost(net) ;
  Real	Wo	   = MstandardWeight(net) ;
  int	idx ;

  if (weightCost == 0.0) return ;
  if (Wo == 0.0) {
    /* WEIGHT DECAY */
    for (idx=0, wt=Mweights(unit), dv=Mderivs(unit) ;
	 idx < numIncoming ;
	 idx++) {
      Real  wtCost = (*wt) * weightCost ;
      net->cost   += wtCost * (*wt++) ;
      (*dv++)     += 2.0 * wtCost ;
    }
  } else {
    /* WEIGHT MELTING */
    for (idx=0, wt=Mweights(unit), dv=Mderivs(unit) ;
	 idx < numIncoming ;
	 idx++) {
      Real ratio = (*wt) * (*wt) / (Wo * Wo) ;
      net->cost += weightCost * ratio / (1.0+ratio) ;
      (*dv++)   += weightCost * 2.0 * (*wt++) / ((1.0+ratio)*(1.0+ratio)) ;
    }
  }
}
/**********************************************************************/
static void	adaptWeightMelting(net)
  Net		net ;
{
  MaveError(net) = (MaveError(net)<0.0 ? net->error :
		    net->error + MaveErrorFactor(net)*(MaveError(net)-net->error)) ;
  if (net->error < MlastError(net)	||
      net->error < MdesiredError(net)	||
      MlastError(net) < 0.0) 
    MweightCost(net) += MweightCostIncrement(net) ;
  else if (net->error < MaveError(net))
    MweightCost(net) -= MweightCostIncrement(net) ;
  else
    MweightCost(net) *= MweightCostDecrement(net) ;
  MlastError(net) = net->error ;
}
/**********************************************************************/
/**********************************************************************/
/***********************************************************************
 *	Name:		netForward
 *	Description:	
 *		Net	net - the net to act on
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void	netForward (net)
  Net		net ;
{
  if (!Mcompiled(net)) netCompile(net) ;
  netForAllOtherUnits (net, INPUT|BIAS, forward, NULL) ;
  netForAllUnits      (net, OUTPUT, 	incrementError, &(net->error)) ;
  if (MunitCost(net) != 0.0)
    netForAllOtherUnits(net, INPUT|OUTPUT|BIAS, incrementCost, &(net->cost)) ;
}
/**********************************************************************/
static	void	forward(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	numIncoming 	= MnumCompiledIncoming(unit) ;
  Real	totalInput  	= 0.0 ;
  Real	*wt, **preOutput ;
  int	idx ;

  if (Mlesioned(unit)) 
    unit->output = 0.0 ;
  else {
    for (idx=0, wt=Mweights(unit), preOutput=MpreOutputs(unit) ;
	 idx < numIncoming ; 
	 idx++) 
      totalInput += (*wt++) * *(*preOutput++) ;
    unit->output = sigmoid(totalInput) ;
  }
}
/**********************************************************************/
/**********************************************************************/
/***********************************************************************
 *	Name:		netBackward
 *	Description:	
 *		Net	net - the net to act on
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void	netBackward (net)
  Net		net ;
{
  if (!Mcompiled(net)) netCompile(net) ;
  netForAllOtherUnits     (net, INPUT|BIAS,	injectError,	NULL) ;
  if (MunitCost(net) != 0.0)
    netForAllOtherUnits(net, INPUT|OUTPUT|BIAS,  injectCost,	NULL) ;
  netForAllOtherUnitsBack (net, INPUT|BIAS,	backward,	NULL) ;
}
/**********************************************************************/
static	void	backward(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	numIncoming = MnumCompiledIncoming(unit) ;
  Real	inputDeriv ;
  Real 	*wt, *dv, **preOut, **preDv ;
  int	idx ;

  if (Mlesioned(unit)) return ;
  inputDeriv = sigmoidDeriv(unit->output) * unit->outputDeriv ;

  for (idx=0, wt=Mweights(unit), dv=Mderivs(unit), preOut=MpreOutputs(unit),
       preDv=MpreOutputDerivs(unit) ;
       idx < numIncoming ; 
       idx++) {
    *(*preDv++) += inputDeriv *  (*wt++) ;
     (*dv++)    += inputDeriv * *(*preOut++) ;
  }
}
/**********************************************************************/
/**********************************************************************/
/**********************************************************************/
static Proc	updateNetActivity (net)
  Net	net ;
{
  if (!Mcompiled(net)) netCompile(net) ;
  netForAllOtherUnits(net, INPUT|BIAS, getCompiledWeights, NULL) ;
  net->error = MlastError(net) = MaveError(net) = 0.0 ;
  MexampleFrequency(net) = 1.0 ;
  netForward(net) ;
  IDoTrace(&afterExample) ;
}
/**********************************************************************/
static Proc	updateNetGradient (net)
  Net	net ;
{
  if (!Mcompiled(net)) IErrorAbort("net not compiled") ;
  netBackward(net) ;
  netForAllOtherUnits(net, INPUT|BIAS, putCompiledDerivs, NULL) ;
}
/***********************************************************************/
static	Boolean	getExample(net, exampleSet, next) 
  Net		net ;
  ExampleSet	exampleSet ;
  int		next ;
{
  if (MbatchSize(net) < exampleSet->numExamples) {
    /* select random example according to it's probability */
    do {
      MexampleNum(net) = randomInteger(exampleSet->numExamples);
      MexampleFrequency(net) = exampleSet->example[MexampleNum(net)].frequency ;
    } while (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net))) ;
  } else {
    MexampleNum(net) = next ;
    MexampleFrequency(net) = exampleSet->example[next].frequency ;
    if (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net)))
      return FALSE ;       /* skip this example */
  }
  MgetNumber(exampleSet, MexampleNum(net)) ;
  return TRUE ;
}
/***********************************************************************
 *	Name:		calculateNetErrorDeriv
 *	Description:	gradient calculation procedure for backprop net
 *			It processes 'MbatchSize(net)' examples
 *	Parameters:	
 *		Net		net - the net to use
 *		ExampleSet	exampleSet - the examples to use
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void	calculateNetErrorDeriv(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  int numExamples ;

  if (!Mcompiled(net)) netCompile(net) ;
  if (net->currentEpoch==0) MlastError(net) = MaveError(net) = -1.0 ;
  initEpoch(net, MalphaOnset(net)) ;
  net->error = 0.0 ;
  net->cost = 0.0 ;
  netForAllOtherUnits(net, INPUT|BIAS, getCompiledWeights, NULL) ;

  for (numExamples = 0 ; numExamples < MbatchSize(net) ; ++numExamples) {
   if (!getExample(net, exampleSet, numExamples)) continue ;

    netForward(net) ;
    netBackward(net) ;
  }
  if (numExamples <= 0)
    IErrorAbort("calculateNetErrorDeriv: no examples processed") ;

  /* update the cost after everything else is done */
  if (MweightCostIncrement(net) > 0.0) adaptWeightMelting(net) ;
  netForAllUnits(net, ALL, updateCost, net) ;

  netForAllOtherUnits(net, INPUT|BIAS, putCompiledDerivs, NULL) ;
}
/***********************************************************************
 *	Name:		calculateNetError
 *	Description:	error calculation procedure for backprop net
 *			It processes 'MbatchSize(net)' examples
 *	Parameters:	
 *		Net		net - the net to use
 *		ExampleSet	exampleSet - the examples to use
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void	calculateNetError(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  int numExamples ;

  if (!Mcompiled(net)) netCompile(net) ;
  if (net->currentEpoch==0) MlastError(net) = MaveError(net) = -1.0 ;
  net->error = 0.0 ;
  net->cost = 0.0 ;
  netForAllOtherUnits(net, INPUT|BIAS, getCompiledWeights, NULL) ;

  for (numExamples = 0 ; numExamples < MbatchSize(net) ; ++numExamples) {
    if (!getExample(net, exampleSet, numExamples)) continue ;

    netForward(net) ;
  }
  if (numExamples <= 0)
    IErrorAbort("calculateNetError: no examples processed") ;

  if (MweightCostIncrement(net) > 0.0) adaptWeightMelting(net) ;
  netForAllUnits(net, ALL, updateCost, net) ;
}
